#!/usr/bin/env python
# encoding: utf-8
#
# @Author: José Sánchez-Gallego
# @Date: Oct 11, 2017
# @Filename: logger.py
# @License: BSD 3-Clause
# @Copyright: José Sánchez-Gallego

# Adapted from astropy's logging system.


from __future__ import absolute_import, division, print_function

import datetime
import logging
import os
import re
import shutil
import sys
import traceback
import warnings
from logging.handlers import TimedRotatingFileHandler

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_by_name

from .color_print import color_text


# from textwrap import TextWrapper




# Adds custom log level for print and twisted messages
PRINT = 15
logging.addLevelName(PRINT, 'PRINT')


def print_log_level(self, message, *args, **kws):
    self._log(PRINT, message, args, **kws)


logging.Logger._print = print_log_level


def print_exception_formatted(type, value, tb):
    """A custom hook for printing tracebacks with colours."""

    tbtext = ''.join(traceback.format_exception(type, value, tb))
    lexer = get_lexer_by_name('pytb', stripall=True)
    formatter = TerminalFormatter()
    sys.stderr.write(highlight(tbtext, lexer, formatter))


def colored_formatter(record):
    """Prints log messages with colours."""

    colours = {'info': ('blue', 'normal'),
               'debug': ('magenta', 'normal'),
               'warning': ('yellow', 'normal'),
               'print': ('green', 'normal'),
               'error': ('red', 'bold')}

    levelname = record.levelname.lower()

    if levelname == 'error':
        return

    if levelname.lower() in colours:
        levelname_color = colours[levelname][0]
        header = color_text('[{}]: '.format(levelname.upper()), levelname_color)

    message = '{0}'.format(record.msg)

    warning_category = re.match(r'^(\w+Warning:).*', message)
    if warning_category is not None:
        warning_category_colour = color_text(warning_category.groups()[0], 'cyan')
        message = message.replace(warning_category.groups()[0], warning_category_colour)

    sub_level = re.match(r'(\[.+\]:)(.*)', message)
    if sub_level is not None:
        sub_level_name = color_text(sub_level.groups()[0], 'red')
        message = '{}{}'.format(sub_level_name, ''.join(sub_level.groups()[1:]))

    # if len(message) > 79:
    #     tw = TextWrapper()
    #     tw.width = 79
    #     tw.subsequent_indent = ' ' * (len(record.levelname) + 2)
    #     tw.break_on_hyphens = False
    #     message = '\n'.join(tw.wrap(message))

    sys.__stdout__.write('{}{}\n'.format(header, message))
    sys.__stdout__.flush()

    return


class MyFormatter(logging.Formatter):

    warning_fmp = '%(asctime)s - %(levelname)s: %(message)s [%(origin)s]'
    info_fmt = '%(asctime)s - %(levelname)s - %(message)s [%(funcName)s @ ' + \
        '%(filename)s]'

    ansi_escape = re.compile(r'\x1b[^m]*m')

    def __init__(self, fmt='%(levelname)s - %(message)s [%(funcName)s @ ' +
                 '%(filename)s]'):
        logging.Formatter.__init__(self, fmt, datefmt='%Y-%m-%d %H:%M:%S')

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.getLevelName('PRINT'):
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.INFO:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.ERROR:
            self._fmt = MyFormatter.info_fmt

        elif record.levelno == logging.WARNING:
            self._fmt = MyFormatter.warning_fmp

        record.msg = self.ansi_escape.sub('', record.msg)

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._fmt = format_orig

        return result


Logger = logging.getLoggerClass()
fmt = MyFormatter()


class LoggerStdout(object):
    """A pipe for stdout to a logger."""

    def __init__(self, level):
        self.level = level

    def write(self, message):

        if message != '\n':
            self.level(message)

    def flush(self):
        pass


class MyLogger(Logger):
    """This class is used to set up the logging system.

    The main functionality added by this class over the built-in
    logging.Logger class is the ability to keep track of the origin of the
    messages, the ability to enable logging of warnings.warn calls and
    exceptions, and the addition of colorized output and context managers to
    easily capture messages to a file or list.

    It is addapted from the astropy logging system.

    """

    INFO = 15

    # The default actor to log to. It is set by the set_actor() method.
    _actor = None

    def save_log(self, path):
        shutil.copyfile(self.log_filename, os.path.expanduser(path))

    def _show_warning(self, *args, **kwargs):

        warning = args[0]
        message = '{0}: {1}'.format(warning.__class__.__name__, args[0])
        mod_path = args[2]

        mod_name = None
        mod_path, ext = os.path.splitext(mod_path)
        for name, mod in sys.modules.items():
            path = os.path.splitext(getattr(mod, '__file__', ''))[0]
            if path == mod_path:
                mod_name = mod.__name__
                break

        if mod_name is not None:
            self.warning(message, extra={'origin': mod_name})
        else:
            self.warning(message)

    def _catch_exceptions(self, exctype, value, tb):
        """Catches all exceptions and logs them."""

        # Now we log it.
        self.error('Uncaught exception', exc_info=(exctype, value, tb))

        # First, we print to stdout with some colouring.
        print_exception_formatted(exctype, value, tb)

    def _set_defaults(self, log_level=logging.INFO, redirect_stdout=False):
        """Reset logger to its initial state."""

        # Remove all previous handlers
        for handler in self.handlers[:]:
            self.removeHandler(handler)

        # Set levels
        self.setLevel(logging.DEBUG)

        # Set up the stdout handler
        self.fh = None
        self.sh = logging.StreamHandler()
        self.sh.emit = colored_formatter
        self.addHandler(self.sh)

        self.sh.setLevel(log_level)

        # warnings.showwarning = self._show_warning

        # Redirects all stdout to the logger
        if redirect_stdout:
            sys.stdout = LoggerStdout(self._print)

        # Catches exceptions
        sys.excepthook = self._catch_exceptions

    def start_file_logger(self, name, log_file_level=logging.DEBUG, log_file_path='./'):
        """Start file logging."""

        log_file_path = os.path.expanduser(log_file_path) / '{}.log'.format(name)
        logdir = log_file_path.parent

        try:
            logdir.mkdir(parents=True, exist_ok=True)

            # If the log file exists, backs it up before creating a new file handler
            if log_file_path.exists():
                strtime = datetime.datetime.utcnow().strftime('%Y-%m-%d_%H:%M:%S')
                shutil.move(log_file_path, log_file_path + '.' + strtime)

            self.fh = TimedRotatingFileHandler(str(log_file_path), when='midnight', utc=True)
            self.fh.suffix = '%Y-%m-%d_%H:%M:%S'
        except (IOError, OSError) as ee:
            warnings.warn('log file {0!r} could not be opened for writing: '
                          '{1}'.format(log_file_path, ee), RuntimeWarning)
        else:
            self.fh.setFormatter(fmt)
            self.addHandler(self.fh)
            self.fh.setLevel(log_file_level)

        self.log_filename = log_file_path


logging.setLoggerClass(MyLogger)
log = logging.getLogger(__name__)
log._set_defaults()  # Inits sh handler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-07-16
# @Filename: test_aperture.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-08 19:41:10

import numpy
import pytest

import marvin
from marvin.tests.conftest import set_the_config


# Inputs are [class, plateifu, release, coords,
#             aperture_params, aperture_type, coord_type, threshold]
# Outputs are [n_pixels, n_spaxels]. n_pixels are pixels on the mask that have
# a value of 1. n_spaxels are the number of spaxels extracted given the threshold.
test_params = [

    [['Cube', '8485-1901', 'MPL-6', (17, 17), 3, 'circular', 'pixel', 0.5], [21, 25]],
    [['Maps', '8485-1901', 'MPL-6', (17, 17), 3, 'circular', 'pixel', 0.5], [21, 25]],
    [['ModelCube', '8485-1901', 'MPL-6', (17, 17), 3, 'circular', 'pixel', 0.5], [21, 25]],

    [['Maps', '8485-1901', 'MPL-6', (232.5447, 48.690201), 1, 'circular', 'sky', 0.8], [5, 9]],

    [['Maps', '8485-1901', 'MPL-6', [(5, 5), (20, 20)], (4, 4, 0), 'rectangular', 'pixel', 0.5],
     [18, 42]],

    [['Maps', '8485-1901', 'MPL-6', (5, 5), (4, 4, 45), 'elliptical', 'pixel', 0.5], [9, 45]],

]


@pytest.mark.parametrize('inputs, outputs', test_params)
def test_get_aperture(inputs, outputs):

    class_name, plateifu, release, coords, \
        aperture_params, aperture_type, coord_type, threshold = inputs

    set_the_config(release)

    aperture_params = numpy.atleast_1d(aperture_params)

    n_pixels_mask, n_spaxels = outputs

    ToolClass = getattr(marvin.tools, class_name)
    instance = ToolClass(plateifu, release=release)

    assert isinstance(instance, ToolClass)
    assert hasattr(instance, 'getAperture')

    aperture = instance.getAperture(coords, aperture_params, aperture_type=aperture_type,
                                    coord_type=coord_type)

    assert isinstance(aperture, marvin.tools.mixins.MarvinAperture)

    if coord_type == 'pixel':
        aperture_positions = aperture.positions
    else:
        aperture_positions = numpy.array([aperture.positions.ra.deg,
                                          aperture.positions.dec.deg]).T

    numpy.testing.assert_allclose(numpy.atleast_2d(coords), aperture_positions)

    if aperture_type == 'circular':
        if coord_type == 'pixel':
            assert aperture.r == pytest.approx(aperture_params[0], rel=1e-6)
        else:
            assert aperture.r.value == pytest.approx(aperture_params[0], rel=1e-6)
    elif aperture_type == 'rectangular':
        if coord_type == 'pixel':
            assert aperture.w == pytest.approx(aperture_params[0], rel=1e-6)
            assert aperture.h == pytest.approx(aperture_params[1], rel=1e-6)
            assert aperture.theta == pytest.approx(aperture_params[2], rel=1e-6)
        else:
            assert aperture.w.value == pytest.approx(aperture_params[0], rel=1e-6)
            assert aperture.h.value == pytest.approx(aperture_params[1], rel=1e-6)
            assert aperture.theta.value == pytest.approx(aperture_params[2], rel=1e-6)
    elif aperture_type == 'elliptical':
        if coord_type == 'pixel':
            assert aperture.a == pytest.approx(aperture_params[0], rel=1e-6)
            assert aperture.b == pytest.approx(aperture_params[1], rel=1e-6)
            assert aperture.theta == pytest.approx(aperture_params[2], rel=1e-6)
        else:
            assert aperture.a.value == pytest.approx(aperture_params[0], rel=1e-6)
            assert aperture.b.value == pytest.approx(aperture_params[1], rel=1e-6)
            assert aperture.theta.value == pytest.approx(aperture_params[2], rel=1e-6)

    assert numpy.sum(aperture.mask == 1) == n_pixels_mask

    spaxels = aperture.getSpaxels(threshold=threshold)
    assert len(spaxels) == n_spaxels

    assert spaxels[0].loaded is False

    # Check that we can load the spaxel
    spaxels[0].load()
    assert spaxels[0].loaded is True


def test_get_aperture_no_lazy():
    """Tests if the lazy=False feature in getSpaxels works."""

    cube = marvin.tools.Cube('8485-1901')

    aperture = cube.getAperture((17, 17), 1)

    assert aperture.getSpaxels(lazy=False)[0].loaded is True


def test_get_spaxels_custom_mask():
    """Tests the mask parameter in getSpaxels."""

    cube = marvin.tools.Cube('8485-1901')

    aperture = cube.getAperture((17, 17), 2)

    my_mask = aperture.mask.copy()
    my_mask[0:5, 0:5] = 1

    assert len(aperture.getSpaxels(mask=my_mask)) == 34



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-25 10:11:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-20 13:29:52

from __future__ import absolute_import, division, print_function

from imp import reload
import pytest

import marvin
from marvin import config
from marvin.core.exceptions import MarvinError
from marvin.tests import marvin_test_if
from marvin.tests.conftest import set_the_config
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.modelcube import ModelCube
from marvin.tools.query import Query, doQuery
from marvin.tools.spaxel import Spaxel


@pytest.fixture(scope='function', autouse=True)
def allow_dap(monkeypatch):
    monkeypatch.setattr(config, '_allow_DAP_queries', True)
    global Query, doQuery
    reload(marvin.tools.query)
    from marvin.tools.query import Query, doQuery


class TestDoQuery(object):

    def test_success(self, release, mode):
        q, r = doQuery(search_filter='nsa.z < 0.1', release=release, mode=mode)
        assert q is not None
        assert r is not None


class TestQueryVersions(object):

    def test_versions(self, query, release, versions):
        drpver, dapver = versions
        assert query.release == release
        assert query._drpver == drpver
        assert query._dapver == dapver


class TestQuerySearches(object):

    @pytest.mark.parametrize('query, joins',
                             [('nsa.z < 0.1', ['nsa', 'drpalias']),
                              ('emline_gflux_ha_6564 > 25', ['spaxelprop', 'dapalias']),
                              ('nsa.z < 0.1 and emline_gflux_ha_6564 > 25', ['nsa', 'spaxelprop', 'drpalias', 'dapalias'])],
                             ids=['drponly', 'daponly', 'drp_and_dap'],
                             indirect=['query'])
    def test_whereclause(self, query, joins):
        if query.mode == 'remote':
            res = query.run()
            for item in joins:
                assert item in str(res.query)
        else:
            for item in joins:
                assert item in str(query.query.whereclause)

    @pytest.mark.parametrize('query, addparam',
                             [('nsa.z < 0.1', ['nsa.z']),
                              ('emline_gflux_ha_6564 > 25', ['emline_gflux_ha_6564', 'spaxelprop.x', 'spaxelprop.y', 'bintype.name', 'template.name'])],
                             indirect=['query'])
    def test_params(self, query, addparam):
        params = query.expdata['defaults'] + addparam
        res = query.run()
        assert set(params) == set(query.params)

    @pytest.mark.parametrize('badquery, errmsg',
                             [('nsa.hello < 0.1', 'nsa.hello does not match any column.'),
                              ('name = SPX', 'name matches multiple parameters in the lookup table'),
                              ('< 0.1', 'Your boolean expression contained a syntax error')],
                             ids=['nomatch', 'multiple_entries', 'syntax_error'])
    def test_bad_queries(self, expmode, badquery, errmsg):
        if expmode is None:
            pytest.skip('cannot use queries in local mode without a db')

        set_the_config(config.release)

        with pytest.raises((KeyError, MarvinError)) as cm:
            query = Query(search_filter=badquery, mode=expmode)
            res = query.run()
        assert cm.type == KeyError or cm.type == MarvinError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('query, sfilter',
                             [('nsa.z < 0.1', 'nsa.z < 0.1'),
                              ('absmag_g_r > -1', 'absmag_g_r > -1'),
                              ('emline_gflux_ha_6564 > 25', 'emline_gflux_ha_6564 > 25'),
                              ('npergood(emline_gflux_ha_6564 > 5) > 20', 'npergood(emline_gflux_ha_6564 > 5) > 20'),
                              ('nsa.z < 0.1 and emline_gflux_ha_6564 > 25', 'nsa.z < 0.1 and emline_gflux_ha_6564 > 25')],
                             indirect=['query'], ids=['nsaz', 'absgr', 'haflux', 'npergood', 'nsahaflux'])
    def test_success_queries(self, query, sfilter):
        res = query.run()
        count = query.expdata['queries'][sfilter]
        assert count['count'] == res.totalcount

    # @pytest.mark.parametrize('query, qmode',
    #                          [('nsa.z < 0.1', 'count'),
    #                           ('nsa.z < 0.1', 'first')],
    #                          indirect=['query'])
    # def test_qmodes(self, query, qmode):
    #     mycount = query.expdata['queries']['nsa.z < 0.1']['count']
    #     r = query.run(qmode)
    #     if qmode == 'count':
    #         assert r == mycount
    #     elif qmode == 'first':
    #         assert len(r.results) == 1
    #         assert r.count == 1


class TestQuerySort(object):

    @pytest.mark.parametrize('query, sortparam, order',
                             [('nsa.z < 0.1', 'z', 'asc'),
                              ('nsa.z < 0.1', 'nsa.z', 'desc')], indirect=['query'])
    def test_sort(self, query, sortparam, order):
        data = query.expdata['queries']['nsa.z < 0.1']['sorted']
        query = Query(search_filter=query.search_filter, mode=query.mode, sort=sortparam, order=order)
        res = query.run()
        if order == 'asc':
            redshift = data['1'][-1]
        else:
            redshift = data['last'][-1]
        assert res.results['z'][0] == redshift


class TestQueryShow(object):

    @pytest.mark.parametrize('query, show, exp',
                             [('nsa.z < 0.1', 'query', 'SELECT mangadatadb.cube.mangaid'),
                              ('nsa.z < 0.1', 'joins', "['ifudesign', 'manga_target', 'manga_target_to_nsa', 'nsa']"),
                              ('nsa.z < 0.1', 'filter', 'mangasampledb.nsa.z < 0.1')], indirect=['query'])
    def test_show(self, query, show, exp, capsys):
        sql = query.show(show)
        if query.mode == 'remote':
            assert sql == query.search_filter
        else:
            assert exp in sql or exp == sql.strip('\n')


# class TestQueryReturnParams(object):

#     @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
#     @pytest.mark.parametrize('rps', [(['g_r']), (['cube.ra', 'cube.dec']), (['haflux'])])
#     def test_success(self, query, rps):
#         query = Query(search_filter=query.search_filter, return_params=rps, mode=query.mode)
#         params = query._remote_params['params'].split(',') if query.mode == 'remote' else query.params
#         #assert 'nsa.z' in params
#         names = [query._marvinform._param_form_lookup._nameShortcuts[r] for r in rps]
#         assert set(names).issubset(set(query.params))
#         res = query.run()
#         assert all([p in res.columns for p in rps]) is True
#         #assert set(rps).issubset(set(query.params))
#         #assert set(rps).issubset(set(res.paramtocol.keys()))

#     @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
#     @pytest.mark.parametrize('rps, errmsg',
#                              [('hello', 'does not match any column.'),
#                               ('name', 'name matches multiple parameters')],
#                              ids=['nomatch', 'multiple_entries'])
#     def test_badparams(self, query, expmode, rps, errmsg):
#         # set error type based on query mode
#         if expmode == 'remote':
#             error = MarvinError
#         else:
#             error = KeyError

#         with pytest.raises(error) as cm:
#             query = Query(search_filter=query.search_filter, return_params=[rps], mode=query.mode)
#             res = query.run()
#         assert cm.type == error
#         assert errmsg in str(cm.value)

#     @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
#     @pytest.mark.parametrize('rps', [(['absmag_g_r', 'cube.plate', 'cube.plateifu'])])
#     def test_skipdefault(self, query, rps):
#         query = Query(search_filter=query.search_filter, return_params=rps, mode=query.mode)
#         params = query._remote_params['params'].split(',') if query.mode == 'remote' else query.params
#         assert len(query.return_params) == len(rps)
#         assert len(params) == 5
#         res = query.run()
#         assert len(res.returnparams) == len(rps)
#         assert len(res.columns) == 5


class TestQueryReturnType(object):

    @pytest.mark.parametrize('query', [('cube.mangaid == 1-209232 and emline_gflux_ha_6564 > 25')], indirect=True)
    @pytest.mark.parametrize('objtype, tool',
                             [('cube', Cube), ('maps', Maps), ('spaxel', Spaxel),
                              ('modelcube', ModelCube)])
    @marvin_test_if(mark='skip', query={'release': ['MPL-4']})
    def test_get_success(self, query, objtype, tool):
        if query.mode == 'remote' and config.db is None:
            pytest.skip('skipping weird case where nodb, remote mode tried to load a local file')
        if config.release == 'MPL-4' and objtype == 'modelcube':
            pytest.skip('no modelcubes in mpl-4')

        query = Query(search_filter=query.search_filter, return_type=objtype, mode=query.mode, release=query.release)
        res = query.run()
        assert res.objects is not None
        assert len(res.results) == len(res.objects)
        assert isinstance(res.objects[0], tool) is True

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('objtype, errmsg',
                             [('noncube', 'Query return_type must be either cube, spaxel, maps, modelcube, rss')])
    @marvin_test_if(mark='skip', query={'mode': ['remote']})
    def test_badreturntype(self, query, objtype, errmsg):
        with pytest.raises(AssertionError) as cm:
            query = Query(search_filter=query.search_filter, return_type=objtype, mode=query.mode)
        assert cm.type == AssertionError
        assert errmsg in str(cm.value)


class TestQueryPickling(object):

    @pytest.mark.parametrize('query', [('nsa.z < 0.1')], indirect=True)
    def test_pickle_save(self, temp_scratch, query):
        if query.mode == 'local':
            pytest.xfail('save cannot be run in local mode')
        file = temp_scratch.join('test_query.mpf')
        query.save(str(file))
        assert file.check() is True

    @pytest.mark.parametrize('query, sfilter', [('nsa.z < 0.1', 'nsa.z < 0.1')], indirect=['query'])
    def test_pickle_restore(self, temp_scratch, query, sfilter):
        if query.mode == 'local':
            pytest.xfail('save cannot be run in local mode')
        file = temp_scratch.join('test_query.mpf')
        query.save(str(file))
        assert file.check() is True
        query = None
        assert query is None
        query = Query.restore(str(file))
        assert query.search_filter == sfilter


@pytest.mark.skip(reason="no spaxel queries causes this to fail since parameter counts are now off")
class TestQueryParams(object):

    @pytest.mark.parametrize('paramdisplay', [('all'), ('best')])
    def test_getparams(self, query, paramdisplay):
        params = query.get_available_params(paramdisplay, release=query.release)
        mydata = query.expdata['params'][paramdisplay]
        # counts and content
        if paramdisplay == 'best':
            assert mydata['count'] == sum([len(v) for v in params])
            assert set(mydata['subset']).issubset(set(params.list_params()))
            assert set(mydata).isdisjoint(set(params.list_params()))
        elif paramdisplay == 'all':
            assert mydata['count'] == len(params)
            assert set(mydata['subset']).issubset(set(params))


class TestQueryModes(object):

    @pytest.mark.parametrize('query', [('nsa.z < 0.1 and cube.plate == 8485')], indirect=True)
    def test_getmode(self, query, expmode):
        assert query.mode == expmode
        res = query.run()
        assert res.mode == expmode
        assert query.mode == res.mode

#
# Below here begins a low refactor of the query tests
#

class TestQueryLocal(object):
    mode = 'local'
    sf = 'nsa.z < 0.1'

    @pytest.mark.parametrize('rps', [(['g_r']), (['cube.ra', 'cube.dec']), (['emline_gflux_ha_6564'])], ids=['g-r', 'radec', 'haflux'])
    def test_return_params(self, rps):
        base = ['cube.ra', 'cube.dec']
        query = Query(search_filter=self.sf, mode=self.mode, return_params=base + rps)
        reals = [query._marvinform._param_form_lookup.get_real_name(r) for r in rps]
        assert set(reals).issubset(set(query.params))
        assert set(reals).issubset(query.return_params)

    @pytest.mark.parametrize('rps, errmsg',
                             [('hello', 'does not match any column.'),
                              ('name', 'name matches multiple parameters')],
                             ids=['nomatch', 'multiple_entries'])
    def test_bad_returnparams(self, rps, errmsg):
        with pytest.raises(KeyError) as cm:
            query = Query(search_filter=self.sf, return_params=[rps], mode=self.mode)
        assert cm.type == KeyError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('rps', [(['absmag_g_r', 'cube.plate', 'cube.plateifu'])])
    def test_skipdefault(self, rps):
        query = Query(search_filter=self.sf, return_params=rps, mode=self.mode)
        assert len(query.return_params) == len(rps)
        assert len(query.params) == 5
        assert query.params.count('cube.plateifu') == 1

    @pytest.mark.parametrize('sf, rps, schema, tables, check',
                             [(None, None, 'sampledb', None, True),
                              (None, None, 'dapdb', None, False),
                              (None, 'stellar_vel', 'dapdb', None, True),
                              ('stellar_vel > 50', None, 'dapdb', 'spaxelprop', True),
                              (None, 'sb_1re', None, 'dapall', True),
                              ('stellar_z < 0.1', None, 'dapdb', 'dapall', True),
                              ('sb_1re > 0.2', None, 'dapdb', 'cube', False)
                              ],
                             ids=['nsaz', 'nodap', 'daprp', 'dap', 'dapallrp',
                                  'dapall', 'wrongtableinschema'])
    def test_checkfor(self, sf, rps, schema, tables, check):
        sf = sf or self.sf
        query = Query(search_filter=sf, return_params=rps, mode=self.mode)
        isin = query._check_for(query.params, schema=schema, tables=tables)
        assert isin == check

    @pytest.mark.parametrize('sf, rps, schema, tables, check',
                             [(None, None, None, ['cube', 'nsa'], True),
                              ('sb_1re > 0.2', None, None, 'dapall', False),
                              ('sb_1re > 0.2', None, 'dapdb', 'dapall', False),
                              ('sb_1re > 0.2', None, 'dapdb', ['dapall', 'bintype', 'template'], True),
                              (None, 'stellar_z', 'dapdb', ['dapall', 'bintype', 'template'], True)
                              ],
                             ids=['drp', 'noschema', 'onlydapall', 'good', 'goodrp'])
    def test_checkonly(self, sf, rps, schema, tables, check):
        sf = sf or self.sf
        query = Query(search_filter=sf, return_params=rps, mode=self.mode)
        isin = query._check_for(query.params, schema=schema, tables=tables, only=True)
        assert isin == check

    @pytest.mark.parametrize('sf, rps',
                             [(None, 'stellar_vel'),
                              ('emline_gflux_ha_6564 > 25', None),
                              ('stellar_z < 0.1 and stellar_vel > 50', None),
                              ('sb_1re > 0.2', 'stellar_vel')],
                             ids=['nodaprp', 'nodap', 'dapall', 'daprp'])
    def test_nodap(self, monkeypatch, sf, rps):
        sf = sf or self.sf
        monkeypatch.setattr(config, '_allow_DAP_queries', False)
        with pytest.raises(NotImplementedError) as cm:
            query = Query(search_filter=sf, return_params=rps, mode=self.mode)
        assert 'DAP spaxel queries are disabled in this version.' in str(cm)

    @pytest.mark.parametrize('sf',
                             [('radial(232.5447, 48.6902, 1)'),
                              ('npergood(emline_gflux_ha_6564 > 5) > 20')],
                             ids=['radial', 'npergood'])
    def test_function_queries(self, sf):
        q = Query(search_filter=sf)
        r = q.run()
        assert r.results is not None

    @pytest.mark.parametrize('sf, qual, flags, count',
                             [(None, None, 'mangadatadb.cube', 5),
                              (None, ['BADFLUX'], 'DRP3QUAL', 0),
                              (None, ['BADFLUX', 'BADZ'], 'DRP3QUAL, DAPQUAL', 0),
                              ('cube.quality = 0', None, 'DRP3QUAL', 5),
                              ('cube.quality & 256', None, 'DRP3QUAL', 0),
                              (None, ['FORESTAR'], 'DAPQUAL', 40),
                              ('cube.quality & ~0', None, 'DRP3QUAL', 0)],
                             ids=['all', 'drp', 'drpdap', 'goodqual',
                                  'badflux', 'forestar', 'not0'])
    def test_quality_queries(self, sf, qual, flags, count):
        q = Query(search_filter=sf, quality=qual, release='MPL-5')
        sql = q.show()
        for flag in flags.split(','):
            assert flag.strip() in sql
        r = q.run()
        assert r.results is not None
        assert r.totalcount == count

    @pytest.mark.parametrize('sf, targ, flags, count',
                             [(None, None, 'mangadatadb.cube', 3282),
                              (None, ['PRIMARY'], 'MNGT%RG1', 846),
                              (None, ['mwa', 'dwarf'], 'MNGT%RG3', 4)],
                             ids=['all', 'primary', 'anc'])
    def test_target_queries(self, sf, targ, flags, count):
        q = Query(search_filter=sf, targets=targ, release='MPL-4')
        sql = q.show()
        for flag in flags.split(','):
            ff = flag.strip()
            assert (ff in sql or ff.replace('%', '%%') in sql)
        r = q.run()
        assert r.results is not None
        assert r.totalcount == count

    @pytest.mark.parametrize('badinput', [('searchfilter'), ('returnparams')])
    def test_bad_input(self, badinput):
        data = 'nsa.z < 0.1' if badinput == 'searchfilter' else ['cube.plate']
        kwargs = {badinput: data}
        with pytest.raises(TypeError) as cm:
            q = Query(**kwargs)
        assert "__init__() got an unexpected keyword argument '{0}'".format(badinput) in str(cm.value)


class TestQueryAuto(object):

    @pytest.mark.parametrize('mode', [('local'), ('remote')])
    def test_mma(self, mode):
        if mode == 'remote':
            config.forceDbOff()
        q = Query()
        assert q.mode == mode


@pytest.fixture(scope='class')
def dboff():
    config.forceDbOff()
    yield True
    config.forceDbOn()


@pytest.mark.usefixtures("dboff")
class TestQueryRemote(object):
    mode = 'remote'
    sf = 'nsa.z < 0.1'

    @pytest.mark.parametrize('rps', [(['g_r']), (['cube.ra', 'cube.dec']),
                                     (['emline_gflux_ha_6564'])],
                             ids=['g-r', 'radec', 'haflux'])
    def test_return_params(self, rps):
        base = ['cube.ra', 'cube.dec']
        query = Query(search_filter=self.sf, mode=self.mode, return_params=base + rps)
        params = query._remote_params['returnparams'].split(',')
        assert set(rps).issubset(set(params))



#!/usr/bin/env python
# encoding: utf-8
#
# conftest.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import


"""
Here you can add fixtures that will be used for all the tests in this
directory. You can also add conftest.py files in underlying subdirectories.
Those conftest.py will only be applies to the tests in that subdirectory and
underlying directories. See https://docs.pytest.org/en/2.7.3/plugins.html for
more information.
"""



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-06-21 17:01:09
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last Modified time: 2018-10-17 00:22:19

from __future__ import absolute_import, division, print_function

import abc
import os
import time

import sdss_access.path
import sdss_access.sync
import six

import marvin
import marvin.tools.plate
from marvin.core.exceptions import MarvinError


__ALL__ = ['VACContainer', 'VACMixIn']


class VACContainer(object):

    def __repr__(self):
        return '<VACContainer ({0})>'.format(', '.join(map(repr, list(self))))

    def __dir__(self):
        props = []
        for value in self.__class__.__dict__.keys():
            if not value.startswith('_'):
                props.append(value)
        return props

    def __getitem__(self, value):
        return getattr(self, value)

    def __iter__(self):
        for value in self.__dir__():
            yield value


class VACMixIn(object, six.with_metaclass(abc.ABCMeta)):
    """MixIn  that allows VAC integration in Marvin.

    This parent class provides common tools for downloading data using
    sdss_access or directly from the sandbox. `~VACMixIn.get_vacs` returns a
    container with properties pointing to all the VACs that subclass from
    `.VACMixIn`. In general, VACs can be added to a class in the following way:

    .. code-block:: python

        from marvin.contrib.vacs.base import VACMixIn

        class Maps(MarvinToolsClass):

            def __init__(self, *args, **kwargs):

                ...

                self.vacs = VACMixIn.get_vacs(self)

    and then the VACs can be accessed as properties in ``my_map.vacs``.

    """

    # The name and description of the VAC.
    name = None
    description = None

    def __init__(self):

        if not sdss_access.sync.RsyncAccess:
            raise MarvinError('sdss_access is not installed')
        else:
            self._release = marvin.config.release
            is_public = 'DR' in self._release
            rsync_release = self._release.lower() if is_public else None
            self.rsync_access = sdss_access.sync.RsyncAccess(public=is_public, release=rsync_release)

    def __repr__(self):
        return '<VAC (name={0}, description={1})>'.format(self.name, self.description)

    @abc.abstractmethod
    def get_data(self, parent_object):
        """Returns VAC data that matches the `parent_object` target.

        This method must be overridden in each subclass of `VACMixIn`. Details
        will depend on the exact implementation and the type of VAC, but in
        general each version of this method must:

        * Check whether the VAC file exists locally.
        * If it does not, download it using `~VACMixIn.download_vac`.
        * Open the file using the appropriate library.
        * Retrieve the VAC data matching ``parent_object``. Usually one will
          use attributes in ``parent_object`` such as ``.mangaid`` or
          ``.plate`` to perform the match.
        * Return the VAC data in whatever format is appropriate.

        """

        pass

    @staticmethod
    def get_vacs(parent_object):
        """Returns a container with all the VACs subclassing from `VACMixIn`.

        Because this method loops over ``VACMixIn.__subclasses__()``, all the
        class that inherit from `VACMixIn` and that must be included in the
        container need to have been imported before calling
        `~VACMixIn.get_vacs`.

        Parameters
        ----------
        parent_object : object
            The object to which the VACs are being attached. It will be passed
            to `~VACMixIn.get_data` when the subclass of `VACMixIn` is
            called.

        Returns
        -------
        vac_container : object
            An instance of a class that contains just a list of properties, one
            for to each on of the VACs that subclass from `VACMixIn`.

        """

        vac_container = VACContainer()

        for subvac in VACMixIn.__subclasses__():

            # Excludes VACs from showing up in Plate
            if issubclass(parent_object.__class__, marvin.tools.plate.Plate):
                continue

            # Only shows VACs if in the include list.
            if (hasattr(subvac, 'include') and subvac.include is not None and
                    not issubclass(parent_object.__class__, subvac.include)):
                continue

            # We need to set sv=subvac in the lambda function to prevent
            # a cell-var-from-loop issue.
            if parent_object._release in subvac.version:
                setattr(VACContainer, subvac.name,
                        property(lambda self, sv=subvac: sv().get_data(parent_object)))

        return vac_container

    def download_vac(self, name=None, path_params={}, verbose=True):
        """Download the VAC using rsync and returns the local path."""

        if name is None:
            name = self.name

        assert name in self.rsync_access.templates, 'VAC path has not been set in the tree.'

        if verbose:
            marvin.log.info('downloading file for VAC {0!r}'.format(self.name))

        self.rsync_access.remote()
        self.rsync_access.add(name, **path_params)
        self.rsync_access.set_stream()
        self.rsync_access.commit()
        paths = self.rsync_access.get_paths()

        # adding a millisecond pause for download to finish and file existence to register
        time.sleep(0.001)

        return paths[0]  # doing this for single files, may need to change

    def get_path(self, name=None, path_params={}):
        """Returns the local VAC path or False if it does not exist."""

        if name is None:
            name = self.name

        # check for and expand any wildcards present in the path_params
        if self.rsync_access.any(name, **path_params):
            files = self.rsync_access.expand(name, **path_params)
            return files[0]
        else:
            return False

    def file_exists(self, name=None, path_params={}):
        """Check whether a file exists."""

        if name is None:
            name = self.name

        if os.path.exists(self.get_path(name=name, path_params=path_params)):
            return True

        return False



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-10-11 17:51:43
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-29 17:23:15

from __future__ import print_function, division, absolute_import

import numpy as np
import astropy
import astropy.units as u
import marvin.tools
from marvin.tools.quantities.spectrum import Spectrum
from marvin.utils.general.general import get_drpall_table
from marvin.utils.plot.scatter import plot as scatplot

from .base import VACMixIn


class HIVAC(VACMixIn):
    """Provides access to the MaNGA-HI VAC.

    VAC name: HI

    URL: link

    Description: Returns HI summary data and spectra

    Authors: David Stark and Karen Masters

    """

    # Required parameters
    name = 'mangahi'
    description = 'Returns HI summary data and spectra'
    version = {'MPL-7': 'v1_0_1',
               'DR15': 'v1_0_1'}

    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube,
               marvin.tools.maps.Maps,
               marvin.tools.modelcube.ModelCube)

    # Required method
    def get_data(self, parent_object):

        # get any parameters you need from the parent object
        plateifu = parent_object.plateifu
        release = parent_object.release

        # define the variables to build a unique path to your VAC file
        path_params = {'ver': self.version[release], 'type': 'all',
                       'plateifu': plateifu, 'program': 'GBT16A_095'}

        # get_path returns False if the files do not exist locally
        allfile = self.get_path('mangahisum', path_params=path_params)
        specfile = self.get_path('mangahispectra', path_params=path_params)

        # download the vac from the SAS if it does not already exist locally
        if not allfile:
            allfile = self.download_vac('mangahisum', path_params=path_params)

        # create container for more complex return data
        hidata = HIData(plateifu, allfile=allfile, specfile=specfile)

        # get the spectral data for that row if it exists
        if hidata._indata and not specfile:
            hidata._specfile = self.download_vac('mangahispectra', path_params=path_params)

        return hidata


class HIData(object):
    ''' A customized class to handle more complex data

    This class handles data from both the HI summary file and the
    individual HI spectral files.  Row data from the summary file
    is returned via the `data` property.  Spectral data can be plotted via
    the `plot_spectrum` method.

    '''

    def __init__(self, plateifu, allfile=None, specfile=None):
        self._allfile = allfile
        self._specfile = specfile
        self._plateifu = plateifu
        self._hi_data = self._open_file(allfile)
        self._indata = plateifu in self._hi_data['plateifu']
        self._specdata = None

    def __repr__(self):
        return 'HI({0})'.format(self._plateifu)

    @staticmethod
    def _open_file(hifile):
        return astropy.io.fits.getdata(hifile, 1)

    @property
    def data(self):
        ''' Returns the FITS row data from the mangaHIall summary file '''

        if not self._indata:
            return "No HI data exists for {0}".format(self._plateifu)

        idx = self._hi_data['plateifu'] == self._plateifu
        return self._hi_data[idx]

    def plot_spectrum(self):
        ''' Plot the HI spectrum '''

        if self._specfile:
            if not self._specdata:
                self._specdata = self._open_file(self._specfile)

            vel = self._specdata['VHI'][0]
            flux = self._specdata['FHI'][0]
            spec = Spectrum(flux, unit=u.Jy, wavelength=vel, wavelength_unit=u.km/u.s)
            ax = spec.plot(ylabel='HI\ Flux', xlabel='Velocity', title=self._plateifu, ytrim='minmax')
            return ax
        return None

    def plot_massfraction(self):
        ''' Plot the HI mass fraction '''

        drpall = get_drpall_table()
        drpall.add_index('plateifu')
        subset = drpall.loc[self._hi_data['plateifu']]
        log_stmass = np.log10(subset['nsa_elpetro_mass'])
        diff = self._hi_data['logMHI'] - log_stmass
        fig, axes = scatplot(log_stmass, diff, with_hist=False, ylim=[-5, 5],
                             xlabel=r'log $M_*$', ylabel=r'log $M_{HI}/M_*$')
        return axes[0]





#!/usr/bin/env python
# encoding: utf-8
#
# test_bpt_pytest.py
#
# Created by José Sánchez-Gallego on 10 Feb 2017.
# Converted to pytest by Brett Andrews on 15 Jun 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import inspect

from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1.axes_divider import LocatableAxes
import numpy as np
import pytest

from marvin.core.exceptions import MarvinError
from marvin.tools.maps import Maps
from marvin.tests import marvin_test_if_class
from marvin.utils.dap.bpt import get_snr, bpt_kewley06
from marvin.core.exceptions import MarvinDeprecationWarning


@pytest.fixture()
def maps(galaxy, mode):
    if galaxy.bintype.name != 'SPX':
        pytest.skip('Only running one bintype for bpt tests')

    # if galaxy.release != 'MPL-6':
    #     pytest.skip('Explicitly skipping here since marvin_test_if_class does not work in 2.7')

    maps = Maps(plateifu=galaxy.plateifu, mode=mode)
    maps.bptsums = galaxy.bptsums if hasattr(galaxy, 'bptsums') else None
    yield maps
    maps = None


@marvin_test_if_class(mark='skip', maps=dict(release=['MPL-4', 'MPL-5']))
class TestBPT(object):

    mechanisms = ['sf', 'comp', 'agn', 'seyfert', 'liner', 'invalid', 'ambiguous']

    @pytest.mark.parametrize('useoi', [(True), (False)], ids=['withoi', 'nooi'])
    def test_bpt(self, maps, useoi):
        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        bptflag = 'nooi' if useoi is False else 'global'

        masks, figure, axes = maps.get_bpt(show_plot=False, return_figure=True, use_oi=useoi)
        assert isinstance(figure, plt.Figure)

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums[bptflag][mech]

        if useoi:
            assert len(axes) == 4
        else:
            assert len(axes) == 3

        for ax in axes:
            assert isinstance(ax, LocatableAxes)

    def test_bpt_diffsn(self, maps):

        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        masks, figure, __ = maps.get_bpt(show_plot=False, return_figure=True,
                                         use_oi=True, snr_min=5)
        assert isinstance(figure, plt.Figure)

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums['snrmin5'][mech]

    def test_bpt_oldsn(self, maps):

        if maps.bptsums is None:
            pytest.skip('no bpt data found in galaxy test data')

        with pytest.warns(MarvinDeprecationWarning) as record:
            masks = maps.get_bpt(snr=5, return_figure=False, show_plot=False)

        assert len(record) == 1
        assert record[0].message.args[0] == ('snr is deprecated. Use snr_min instead. '
                                             'snr will be removed in a future version of marvin')

        for mech in self.mechanisms:
            assert mech in masks.keys()
            assert np.sum(masks[mech]['global']) == maps.bptsums['snrmin5'][mech]

    def test_bind_to_figure(self, maps):

        __, __, axes = maps.get_bpt(show_plot=False, return_figure=True)

        assert len(axes) == 4

        for ax in axes:
            new_fig = ax.bind_to_figure()
            assert isinstance(new_fig, plt.Figure)
            assert new_fig.axes[0].get_ylabel() != ''

    def test_kewley_snr_warning(self, maps):

        with pytest.warns(MarvinDeprecationWarning) as record:
            bpt_kewley06(maps, snr=1, return_figure=False)

        assert len(record) == 1
        assert record[0].message.args[0] == ('snr is deprecated. Use snr_min instead. '
                                             'snr will be removed in a future version of marvin')

    def test_wrong_kw(self, maps):

        with pytest.raises(MarvinError) as error:
            maps.get_bpt(snr=5, return_figure=False, show_plot=False, extra_keyword='hola')

        with pytest.raises(MarvinError) as error2:
            bpt_kewley06(maps, snr=5, return_figure=False, extra_keyword='hola')

        assert str(error.value) == 'unknown keyword extra_keyword'
        assert str(error2.value) == 'unknown keyword extra_keyword'


class TestGetSNR(object):

    def test_get_snr_in_dict(self):

        assert get_snr({'ha': 5, 'hb': 4}, 'ha') == 5

    def test_get_snr_default(self):

        default = inspect.getargspec(get_snr).defaults[0]

        assert get_snr({'ha': 5, 'hb': 4}, 'xx') == default



#!/usr/bin/env python
# encoding: utf-8
#
# test_maskbit.py
#
# @Author: Brett Andrews <andrews>
# @Date:   2017-10-06 10:10:00
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-26 12:08:06


from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
import pytest

from marvin.utils.general.maskbit import Maskbit


bits = [(0, 'BITZERO', 'The zeroth bit.'),
        (1, 'BITONE', 'The first bit.'),
        (2, 'BITTWO', 'The second bit.'),
        (3, 'BITTHREE', 'The third bit'),
        (4, 'BITFOUR', 'The fourth bit')]
schema = pd.DataFrame(bits, columns=['bit', 'label', 'description'])
name = 'MYMASK'
description = 'My first Maskbit.'
mask = np.array([[0, 1], [2, 12]])
custom_mask = np.array([[1, 5], [7, 11]])


class TestMaskbit(object):

    def test_maskbit_init_with_schema(self, name=name, schema=schema, description=description):
        mb = Maskbit(name=name, schema=schema, description=description)
        assert np.all(mb.schema == schema)
        assert mb.name == name
        assert mb.description == description
        assert str(mb) == "<Maskbit 'MYMASK' None>".format(schema)

    @pytest.mark.parametrize('name',
                             ['MANGA_TARGET1',
                              'MANGA_DAPPIXMASK'])
    def test_maskbit_init_from_name(self, name):
        mb = Maskbit(name=name)
        assert mb.name == name
        assert isinstance(mb.schema, pd.DataFrame)
        assert mb.description is None

    def test_values_to_bits_no_value_error(self,  name=name, schema=schema, description=description):
        mb = Maskbit(name=name, schema=schema, description=description)
        with pytest.raises(AssertionError) as ee:
            mb.values_to_bits(values=None)

        assert 'Must provide values.' in str(ee.value)

    @pytest.mark.parametrize('values, expected',
                             [(None, [[[], [0]], [[1], [2, 3]]]),
                              (0, []),
                              (3, [0, 1]),
                              (np.array([1, 3]), [[0], [0, 1]]),
                              (np.array([[0, 2], [1, 5]]), [[[], [1]], [[0], [0, 2]]])])
    def test_values_to_bits(self, values, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        if values is None:
            mb.mask = mask

        actual = mb.values_to_bits(values=values)
        assert actual == expected

    @pytest.mark.parametrize('value, expected',
                             [(0, []),
                              (3, [0, 1])])
    def test_value_to_bits(self, value, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        actual = mb._value_to_bits(value, schema.bit.values)
        assert actual == expected

    @pytest.mark.parametrize('values, expected',
                             [(None, [[[], ['BITZERO']], [['BITONE'], ['BITTWO', 'BITTHREE']]]),
                              (0, []),
                              (3, ['BITZERO', 'BITONE']),
                              (np.array([1, 3]), [['BITZERO'], ['BITZERO', 'BITONE']]),
                              (np.array([[0, 2], [1, 3]]), [[[], ['BITONE']], [['BITZERO'], ['BITZERO', 'BITONE']]])])
    def test_values_to_labels(self, values, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        if values is None:
            mb.mask = mask

        actual = mb.values_to_labels(values=values)
        assert actual == expected

    @pytest.mark.parametrize('bits_in, expected',
                             [([], []),
                              ([0], ['BITZERO']),
                              ([1], ['BITONE']),
                              ([[1]], [['BITONE']]),
                              ([[0, 1], [2]], [['BITZERO', 'BITONE'], ['BITTWO']])])
    def test_bits_to_labels(self, bits_in, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        actual = mb._bits_to_labels(bits_in)
        assert actual == expected

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', 2),
                              (['BITONE'], 2),
                              (['BITONE', 'BITTWO'], 6)])
    def test_labels_to_value(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        actual = mb.labels_to_value(labels)
        assert actual == expected

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', [1]),
                              (['BITONE'], [1]),
                              (['BITONE', 'BITTWO'], [1, 2])])
    def test_labels_to_bits(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        actual = mb.labels_to_bits(labels)
        assert actual == expected

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', np.array([[0, 0], [2, 0]])),
                              (['BITONE'], np.array([[0, 0], [2, 0]])),
                              (['BITTWO', 'BITTHREE'], np.array([[0, 0], [0, 12]]))])
    def test_get_mask_int(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        mb.mask = mask
        actual = mb.get_mask(labels)
        assert (actual == expected).all()

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', np.array([[False, False], [True, False]])),
                              (['BITONE'], np.array([[False, False], [True, False]])),
                              (['BITTWO', 'BITTHREE'], np.array([[False, False], [False, True]]))])
    def test_get_mask_bool(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        mb.mask = mask
        actual = mb.get_mask(labels, dtype=bool)
        assert (actual == expected).all()

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', np.array([[0, 0], [2, 2]])),
                              (['BITONE'], np.array([[0, 0], [2, 2]])),
                              (['BITONE', 'BITTHREE'], np.array([[0, 0], [2, 10]]))])
    def test_get_mask_custom_mask_int(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        mb.mask = mask
        actual = mb.get_mask(labels, mask=custom_mask)
        assert (actual == expected).all()

    @pytest.mark.parametrize('labels, expected',
                             [('BITONE', np.array([[False, False], [True, True]])),
                              (['BITONE'], np.array([[False, False], [True, True]])),
                              (['BITONE', 'BITTHREE'], np.array([[False, False], [True, True]]))])
    def test_get_mask_custom_mask_bool(self, labels, expected):
        mb = Maskbit(name=name, schema=schema, description=description)
        mb.mask = mask
        actual = mb.get_mask(labels, mask=custom_mask, dtype=bool)
        assert (actual == expected).all()

    @pytest.mark.parametrize('labels', ['BITFAKE', ['BITFAKE', 'BITTHREE']])
    def test_get_mask_nonpresent_label(self, labels):

        mb = Maskbit(name=name, schema=schema, description=description)

        with pytest.raises(ValueError) as ee:
            mb.get_mask(labels)

        assert 'label \'BITFAKE\' not found in the maskbit schema.' in str(ee)

    @pytest.mark.parametrize('labels, dtype, expected',
                             [('BITFOUR', bool, np.array([[False, False], [False, False]])),
                              ('BITFOUR', int, np.array([[0, 0], [0, 0]]))])
    def test_get_mask_empty(self, labels, dtype, expected):

        mb = Maskbit(name=name, schema=schema, description=description)
        mb.mask = mask
        actual = mb.get_mask(labels, mask=custom_mask, dtype=dtype)
        assert (actual == expected).all()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-07-09
# @Filename: test_imports.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by:   Brian Cherinka
# @Last modified time: 2018-07-09 12:11:48

import marvin


class TestImports(object):

    def test_access_tools_full_path(self):

        assert marvin.tools.plate.Plate is not None
        assert marvin.tools.cube.Cube is not None
        assert marvin.tools.maps.Maps is not None
        assert marvin.tools.modelcube.ModelCube is not None
        assert marvin.tools.spaxel.Spaxel is not None
        assert marvin.tools.spaxel.Bin is not None
        assert marvin.tools.image.Image is not None

    def test_access_tools_from_root(self):

        assert marvin.tools.Plate is not None
        assert marvin.tools.Cube is not None
        assert marvin.tools.Maps is not None
        assert marvin.tools.ModelCube is not None
        assert marvin.tools.Spaxel is not None
        assert marvin.tools.Bin is not None
        assert marvin.tools.Image is not None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-07-02
# @Filename: test_map.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-08-09 16:08:28


import operator
from copy import deepcopy

import matplotlib
import numpy as np
import pytest
from astropy import units as u

from marvin.core.exceptions import MarvinError
from marvin.tests import marvin_test_if
from marvin.tools.maps import Maps
from marvin.tools.quantities import EnhancedMap, Map
from marvin.utils.datamodel.dap import datamodel
from marvin.utils.general.maskbit import Maskbit


value1 = np.array([[16.35, 0.8],
                   [0, -10.]])
value2 = np.array([[591., 1e-8],
                   [4., 10]])

value_prod12 = np.array([[9.66285000e+03, 8e-9],
                         [0, -100]])

value_log2 = np.array([[2.77158748, -8.],
                       [0.60205999, 1.]])

ivar1 = np.array([[4, 1],
                  [6.97789734e+36, 1e8]])
ivar2 = np.array([[10, 1e-8],
                  [5.76744385e+36, 0]])

ivar_sum12 = np.array([[2.85714286e+00, 9.99999990e-09],
                       [3.15759543e+36, 0]])

ivar_prod12 = np.array([[1.10616234e-05, 1.56250000e-08],
                        [0, 0.]])

ivar_pow_2 = np.array([[5.23472002e-08, 9.53674316e-01],
                      [0, 25]])
ivar_pow_05 = np.array([[3.66072168e-03, 7.81250000e+00],
                       [0, 0]])
ivar_pow_0 = np.array([[0, 0],
                      [0, 0]])
ivar_pow_m1 = np.array([[4, 1.],
                        [0, 1e+08]])
ivar_pow_m2 = np.array([[2.67322500e+02, 1.6e-01],
                        [0, 2.5e+09]])
ivar_pow_m05 = np.array([[0.97859327, 5],
                         [0, 0]])

ivar_log1 = np.array([[3.67423420e-04, 4.34294482e+07],
                      [4.11019127e-20, 4.34294482e-06]])

u_flux = u.erg / u.cm**2 / u.s / u.def_unit('spaxel')
u_flux2 = u_flux * u_flux

ufuncs = [it for it in dir(np) if isinstance(getattr(np, it), np.ufunc)]


def _get_maps_kwargs(galaxy, data_origin):

    if data_origin == 'file':
        maps_kwargs = dict(filename=galaxy.mapspath)
    else:
        maps_kwargs = dict(plateifu=galaxy.plateifu, release=galaxy.release,
                           bintype=galaxy.bintype, template_kin=galaxy.template,
                           mode='local' if data_origin == 'db' else 'remote')

    return maps_kwargs


@pytest.fixture(scope='function', params=[('emline_gflux', 'ha_6564'),
                                          ('emline_gvel', 'oiii_5008'),
                                          ('stellar_vel', None),
                                          ('stellar_sigma', None)])
def map_(request, galaxy, data_origin):
    maps = Maps(**_get_maps_kwargs(galaxy, data_origin))
    map_ = maps.getMap(property_name=request.param[0], channel=request.param[1])
    map_.data_origin = data_origin
    return map_


class TestMap(object):

    def test_map(self, map_, galaxy):

        assert map_.getMaps().release == galaxy.release

        assert tuple(map_.shape) == tuple(galaxy.shape)
        assert map_.value.shape == tuple(galaxy.shape)
        assert map_.ivar.shape == tuple(galaxy.shape)
        assert map_.mask.shape == tuple(galaxy.shape)

        assert (map_.masked.data == map_.value).all()
        assert (map_.masked.mask == map_.mask.astype(bool)).all()

        assert map_.snr == pytest.approx(np.abs(map_.value * np.sqrt(map_.ivar)))

        assert datamodel[map_.getMaps()._dapver][map_.datamodel.full()].unit == map_.unit

    def test_plot(self, map_):
        fig, ax = map_.plot()
        assert isinstance(fig, matplotlib.figure.Figure)
        assert isinstance(ax, matplotlib.axes._subplots.Subplot)
        assert 'Make single panel map or one panel of multi-panel map plot.' in map_.plot.__doc__

    @marvin_test_if(mark='skip', map_={'data_origin': ['db']})
    def test_save_and_restore(self, temp_scratch, map_):

        fout = temp_scratch.join('test_map.mpf')
        map_.save(str(fout))
        assert fout.check() is True

        map_restored = Map.restore(str(fout), delete=True)
        assert tuple(map_.shape) == tuple(map_restored.shape)

    @pytest.mark.parametrize('property_name, channel',
                             [('emline_gflux', 'ha_6564'),
                              ('stellar_vel', None)])
    def test_deepcopy(self, galaxy, property_name, channel):
        maps = Maps(plateifu=galaxy.plateifu)
        map1 = maps.getMap(property_name=property_name, channel=channel)
        map2 = deepcopy(map1)

        for attr in vars(map1):
            if not attr.startswith('_'):
                value = getattr(map1, attr)
                value2 = getattr(map2, attr)

                if isinstance(value, np.ndarray):
                    assert np.isclose(value, value2).all()

                elif isinstance(value, np.ma.core.MaskedArray):
                    assert (np.isclose(value.data, value2.data).all() and
                            (value.mask == value2.mask).all())

                elif isinstance(value, Maskbit) or isinstance(value[0], Maskbit):

                    if isinstance(value, Maskbit):
                        value = [value]
                        value2 = [value2]

                    for mb, mb2 in zip(value, value2):
                        for it in ['bits', 'description', 'labels', 'mask', 'name']:
                            assert getattr(mb, it) == getattr(mb2, it)

                        assert (mb.schema == mb2.schema).all().all()

                elif isinstance(value, Maps):
                    pass

                else:
                    assert value == value2, attr

    def test_getMap_invalid_property(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        with pytest.raises(ValueError) as ee:
            maps.getMap(property_name='mythical_property')

        assert 'Your input value is too ambiguous.' in str(ee.value)

    def test_getMap_invalid_channel(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        with pytest.raises(ValueError) as ee:
            maps.getMap(property_name='emline_gflux', channel='mythical_channel')

        assert 'Your input value is too ambiguous.' in str(ee.value)

    @marvin_test_if(mark='include', maps={'plateifu': '8485-1901',
                                          'release': 'MPL-6',
                                          'mode': 'local',
                                          'data_origin': 'file'})
    def test_quatities_reorder(self, maps):
        """Asserts the unit survives a quantity reorder (issue #374)."""

        ha = maps['emline_gflux_ha']

        assert ha is not None
        assert ha.unit is not None

        reordered_ha = np.moveaxis(ha, 0, -1)
        assert reordered_ha.unit is not None

    @marvin_test_if(mark='include', maps={'plateifu': '8485-1901',
                                          'release': 'MPL-6',
                                          'bintype': ['SPX']})
    def test_get_spaxel(self, maps):
        """Tests `.Map.getSpaxel`."""

        ha = maps['emline_gflux_ha']

        spaxel = ha.getSpaxel(x=10, y=10, xyorig='lower')

        assert spaxel is not None
        assert spaxel.x == 10 and spaxel.y == 10

    @marvin_test_if(mark='skip', galaxy=dict(release=['MPL-6']))
    def test_stellar_sigma_values(self, maps, galaxy):
        ''' Assert values for stellar_sigma and stellar_sigmacorr are different (issue #411) '''

        ss = maps.stellar_sigma
        sc = maps.stellar_sigmacorr
        compare = sum(ss == sc)
        assert len(np.unique(compare)) > 1
        x = galaxy.dap['x']
        y = galaxy.dap['y']
        ssvalue = galaxy.dap['stellar_sigma'][galaxy.bintype.name]
        scvalue = galaxy.dap['stellar_sigmacorr'][galaxy.bintype.name]
        assert ssvalue == pytest.approx(ss[x, y].value, 1e-4)
        assert scvalue == pytest.approx(sc[x, y].value, 1e-4)

    def test_datamodel(self, maps):

        gew_ha = maps.emline_gew_ha_6564
        assert gew_ha.datamodel.description == ('Gaussian-fitted equivalent widths measurements '
                                                '(based on EMLINE_GFLUX). Channel = H-alpha 6564.')
    @marvin_test_if(mark='include', galaxy=dict(release=['MPL-6']))
    def test_stellar_sigma_mpl6(self, maps, galaxy):
        with pytest.raises(MarvinError) as cm:
            __ = maps.stellar_sigmacorr
        assert 'stellar_sigmacorr is unreliable in MPL-6. Please use MPL-7.' in str(cm.value)


class TestMapArith(object):

    def test_add_constant(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']
        ha10 = ha + 10.

        assert ha10.value == pytest.approx(ha.value + 10.)
        assert ha10.ivar == pytest.approx(ha.ivar)
        assert ha10.mask == pytest.approx(ha.mask)

    def test_subtract_constant(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']
        ha10 = ha - 10.

        assert ha10.value == pytest.approx(ha.value - 10.)
        assert ha10.ivar == pytest.approx(ha.ivar)
        assert ha10.mask == pytest.approx(ha.mask)

    def test_multiply_constant(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']
        ha10 = ha * 10.

        assert ha10.value == pytest.approx(ha.value * 10.)
        assert ha10.ivar == pytest.approx(ha.ivar / 10.**2)
        assert ha10.mask == pytest.approx(ha.mask)

    def test_divide_constant(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']
        ha10 = ha / 10.

        assert ha10.value == pytest.approx(ha.value / 10.)
        assert ha10.ivar == pytest.approx(ha.ivar * 10.**2)
        assert ha10.mask == pytest.approx(ha.mask)

    @pytest.mark.parametrize('ivar1, ivar2, expected',
                             [(ivar1, ivar2, ivar_sum12)])
    def test_add_ivar(self, ivar1, ivar2, expected):
        assert Map._add_ivar(ivar1, ivar2) == pytest.approx(expected)

    @pytest.mark.parametrize('ivar1, ivar2, value1, value2, value_prod12, expected',
                             [(ivar1, ivar2, value1, value2, value_prod12, ivar_prod12)])
    def test_mul_ivar(self, ivar1, ivar2, value1, value2, value_prod12, expected):
        ivar = Map._mul_ivar(ivar1, ivar2, value1, value2, value_prod12)
        ivar[np.isnan(ivar)] = 0
        ivar[np.isinf(ivar)] = 0
        assert ivar == pytest.approx(expected)

    @pytest.mark.parametrize('power, expected',
                             [(2, ivar_pow_2),
                              (0.5, ivar_pow_05),
                              (0, ivar_pow_0),
                              (-1, ivar_pow_m1),
                              (-2, ivar_pow_m2),
                              (-0.5, ivar_pow_m05)])
    @pytest.mark.parametrize('ivar, value,',
                             [(ivar1, value1)])
    def test_pow_ivar(self, ivar, value, power, expected):
        ivar = Map._pow_ivar(ivar, value, power)
        ivar[np.isnan(ivar)] = 0
        ivar[np.isinf(ivar)] = 0
        assert ivar == pytest.approx(expected)

    @pytest.mark.parametrize('power', [2, 0.5, 0, -1, -2, -0.5])
    def test_pow_ivar_none(self, power):
        assert Map._pow_ivar(None, np.arange(4), power) == pytest.approx(np.zeros(4))

    @pytest.mark.parametrize('ivar, value, expected',
                             [(ivar1, value2, ivar_log1)])
    def test_log10_ivar(self, ivar, value, expected):
        actual = Map._log10_ivar(ivar, value)
        assert actual == pytest.approx(expected)

    def test_log10(self, maps_release_only):
        niiha = maps_release_only.emline_gflux_nii_6585 / maps_release_only.emline_gflux_nii_6585
        log_niiha = np.log10(niiha)
        ivar = np.log10(np.e) * niiha.ivar**-0.5 / niiha.value

        assert log_niiha.value == pytest.approx(np.log10(niiha.value), nan_ok=True)
        assert log_niiha.ivar == pytest.approx(ivar, nan_ok=True)
        assert (log_niiha.mask == niiha.mask).all()
        assert log_niiha.unit == u.dimensionless_unscaled

    @pytest.mark.runslow
    @marvin_test_if(mark='skip', ufunc='log10')
    @pytest.mark.parametrize('ufunc', ufuncs)
    def test_np_ufunc_notimplemented(self, maps_release_only, ufunc):
        ha = maps_release_only.emline_gflux_ha_6564
        nii = maps_release_only.emline_gflux_nii_6585

        with pytest.raises(NotImplementedError) as ee:
            if getattr(getattr(np, ufunc), 'nargs') <= 2:
                getattr(np, ufunc)(ha)

            else:
                getattr(np, ufunc)(nii, ha)

        expected = 'np.{0} is not implemented for Map.'.format(getattr(np, ufunc).__name__)
        assert str(ee.value) == expected

    @pytest.mark.parametrize('unit1, unit2, op, expected',
                             [(u_flux, u_flux, '+', u_flux),
                              (u_flux, u_flux, '-', u_flux),
                              (u_flux, u_flux, '*', u_flux2),
                              (u_flux, u_flux, '/', u.dimensionless_unscaled),
                              (u.km, u.s, '*', u.km * u.s),
                              (u.km, u.s, '/', u.km / u.s)])
    def test_unit_propagation(self, unit1, unit2, op, expected):
        assert Map._unit_propagation(unit1, unit2, op) == expected

    @pytest.mark.parametrize('unit1, unit2, op',
                             [(u_flux, u.km, '+'),
                              (u_flux, u.km, '-')])
    def test_unit_propagation_mismatch(self, unit1, unit2, op):
        with pytest.warns(UserWarning):
            assert Map._unit_propagation(unit1, unit2, op) is None

    @pytest.mark.parametrize('property1, channel1, property2, channel2',
                             [('emline_gflux', 'ha_6564', 'emline_gflux', 'nii_6585'),
                              ('emline_gvel', 'ha_6564', 'stellar_vel', None)])
    def test_add_maps(self, galaxy, property1, channel1, property2, channel2):
        maps = Maps(plateifu=galaxy.plateifu)
        map1 = maps.getMap(property_name=property1, channel=channel1)
        map2 = maps.getMap(property_name=property2, channel=channel2)
        map12 = map1 + map2

        assert map12.value == pytest.approx(map1.value + map2.value)
        assert map12.ivar == pytest.approx(map1._add_ivar(map1.ivar, map2.ivar))
        assert map12.mask == pytest.approx(map1.mask | map2.mask)

    @pytest.mark.parametrize('property1, channel1, property2, channel2',
                             [('emline_gflux', 'ha_6564', 'emline_gflux', 'nii_6585'),
                              ('emline_gvel', 'ha_6564', 'stellar_vel', None)])
    def test_subtract_maps(self, galaxy, property1, channel1, property2, channel2):
        maps = Maps(plateifu=galaxy.plateifu)
        map1 = maps.getMap(property_name=property1, channel=channel1)
        map2 = maps.getMap(property_name=property2, channel=channel2)
        map12 = map1 - map2

        assert map12.value == pytest.approx(map1.value - map2.value)
        assert map12.ivar == pytest.approx(map1._add_ivar(map1.ivar, map2.ivar))
        assert map12.mask == pytest.approx(map1.mask | map2.mask)

    @pytest.mark.parametrize('property1, channel1, property2, channel2',
                             [('emline_gflux', 'ha_6564', 'emline_gflux', 'nii_6585'),
                              ('emline_gvel', 'ha_6564', 'stellar_vel', None)])
    def test_multiply_maps(self, galaxy, property1, channel1, property2, channel2):
        maps = Maps(plateifu=galaxy.plateifu)
        map1 = maps.getMap(property_name=property1, channel=channel1)
        map2 = maps.getMap(property_name=property2, channel=channel2)
        map12 = map1 * map2

        ivar = map1._mul_ivar(map1.ivar, map2.ivar, map1.value, map2.value, map12.value)
        ivar[np.isnan(ivar)] = 0
        ivar[np.isinf(ivar)] = 0

        assert map12.value == pytest.approx(map1.value * map2.value)
        assert map12.ivar == pytest.approx(ivar)
        assert map12.mask == pytest.approx(map1.mask | map2.mask)

    @pytest.mark.parametrize('property1, channel1, property2, channel2',
                             [('emline_gflux', 'ha_6564', 'emline_gflux', 'nii_6585'),
                              ('emline_gvel', 'ha_6564', 'stellar_vel', None)])
    def test_divide_maps(self, galaxy, property1, channel1, property2, channel2):
        maps = Maps(plateifu=galaxy.plateifu)
        map1 = maps.getMap(property_name=property1, channel=channel1)
        map2 = maps.getMap(property_name=property2, channel=channel2)
        map12 = map1 / map2

        ivar = map1._mul_ivar(map1.ivar, map2.ivar, map1.value, map2.value, map12.value)
        ivar[np.isnan(ivar)] = 0
        ivar[np.isinf(ivar)] = 0

        mask = map1.mask | map2.mask
        bad = np.isnan(map12.value) | np.isinf(map12.value)
        mask[bad] = mask[bad] | map12.pixmask.labels_to_value('DONOTUSE')

        with np.errstate(divide='ignore', invalid='ignore'):
            assert map12.value == pytest.approx(map1.value / map2.value, nan_ok=True)

        assert map12.ivar == pytest.approx(ivar)
        assert map12.mask == pytest.approx(mask)

    @pytest.mark.runslow
    @pytest.mark.parametrize('power', [2, 0.5, 0, -1, -2, -0.5])
    @pytest.mark.parametrize('property_name, channel',
                             [('emline_gflux', 'ha_6564'),
                              ('stellar_vel', None)])
    def test_pow(self, galaxy, property_name, channel, power):
        maps = Maps(plateifu=galaxy.plateifu)
        map_orig = maps.getMap(property_name=property_name, channel=channel)
        map_new = map_orig**power

        sig_orig = np.sqrt(1. / map_orig.ivar)
        sig_new = map_new.value * power * sig_orig * map_orig.value
        ivar_new = 1 / sig_new**2.
        ivar_new[np.isnan(ivar_new)] = 0
        ivar_new[np.isinf(ivar_new)] = 0

        assert map_new.value == pytest.approx(map_orig.value**power, nan_ok=True)
        assert map_new.ivar == pytest.approx(ivar_new)
        assert (map_new.mask == map_orig.mask).all()

    @marvin_test_if(mark='skip', galaxy=dict(release=['MPL-4', 'MPL-6']))
    def test_stellar_sigma_correction(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        stsig = maps['stellar_sigma']
        stsigcorr = maps['stellar_sigmacorr']

        expected = (stsig**2 - stsigcorr**2)**0.5
        expected.ivar = (expected.value / stsig.value) * stsig.ivar
        expected.ivar[stsig.ivar == 0] = 0
        expected.ivar[stsigcorr.value >= stsig.value] = 0
        expected.value[stsigcorr.value >= stsig.value] = 0

        actual = stsig.inst_sigma_correction()

        assert actual.value == pytest.approx(expected.value, nan_ok=True)
        assert actual.ivar == pytest.approx(expected.ivar)
        assert (actual.mask == expected.mask).all()
        assert actual.datamodel == stsig.datamodel

    @marvin_test_if(mark='include', galaxy=dict(release=['MPL-4', 'MPL-6']))
    def test_stellar_sigma_correction_MPL4(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        stsig = maps['stellar_sigma']

        if galaxy.release == 'MPL-4':
            errmsg = 'Instrumental broadening correction not implemented for MPL-4.'
        elif galaxy.release == 'MPL-6':
            errmsg = 'The stellar sigma corrections in MPL-6 are unreliable. Please use MPL-7.'

        with pytest.raises(MarvinError) as ee:
            stsig.inst_sigma_correction()

        assert errmsg in str(ee.value)

    def test_stellar_sigma_correction_invalid_property(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']

        with pytest.raises(MarvinError) as ee:
            ha.inst_sigma_correction()

        assert ('Cannot correct {0}_{1} '.format(ha.datamodel.name, ha.datamodel.channel) +
                'for instrumental broadening.') in str(ee.value)

    def test_emline_sigma_correction(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        hasig = maps['emline_gsigma_ha_6564']
        emsigcorr = maps['emline_instsigma_ha_6564']

        expected = (hasig**2 - emsigcorr**2)**0.5
        expected.ivar = (expected.value / hasig.value) * hasig.ivar
        expected.ivar[hasig.ivar == 0] = 0
        expected.ivar[emsigcorr.value >= hasig.value] = 0
        expected.value[emsigcorr.value >= hasig.value] = 0

        actual = hasig.inst_sigma_correction()

        assert actual.value == pytest.approx(expected.value, nan_ok=True)
        assert actual.ivar == pytest.approx(expected.ivar)
        assert (actual.mask == expected.mask).all()
        assert actual.datamodel == hasig.datamodel

    @marvin_test_if(mark='skip', galaxy=dict(release=['MPL-4', 'MPL-5']))
    @pytest.mark.parametrize('channel, op',
                             [('hb', '*'),
                              ('d4000', '*'),
                              ('cn1', '+'),
                              ])
    def test_specindex_sigma_correction(self, galaxy, channel, op):
        maps = Maps(plateifu=galaxy.plateifu)
        si = maps['specindex_' + channel]
        sicorr = maps['specindex_corr' + channel]

        ops = {'+': operator.add, '-': operator.sub, '*': operator.mul, '/': operator.truediv}
        expected = ops[op](si, sicorr)

        actual = si.specindex_correction()

        assert actual.value == pytest.approx(expected.value, nan_ok=True)
        assert actual.ivar == pytest.approx(expected.ivar)
        assert (actual.mask == expected.mask).all()
        assert actual.datamodel == si.datamodel


class TestMaskbit(object):

    def test_masked(self, maps_release_only):
        params = maps_release_only.datamodel.parent.get_default_plot_params()
        ha = maps_release_only['emline_gflux_ha_6564']
        expected = ha.pixmask.get_mask(params['default']['bitmasks'], dtype=bool)

        assert ha.masked.data == pytest.approx(ha.value)
        assert (ha.masked.mask == expected).all()

    @marvin_test_if(mark='include', maps_release_only=dict(release=['MPL-4']))
    def test_values_to_bits_mpl4(self, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.values_to_bits(1) == [0]

    @marvin_test_if(mark='skip', maps_release_only=dict(release=['MPL-4']))
    def test_values_to_bits(self, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.values_to_bits(3) == [0, 1]

    @marvin_test_if(mark='include', maps_release_only=dict(release=['MPL-4']))
    def test_values_to_labels_mpl4(self, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.values_to_labels(1) == ['DONOTUSE']

    @marvin_test_if(mark='skip', maps_release_only=dict(release=['MPL-4']))
    def test_values_to_labels(self, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.values_to_labels(3) == ['NOCOV', 'LOWCOV']

    @marvin_test_if(mark='include', maps_release_only=dict(release=['MPL-4']))
    def test_labels_to_value_mpl4(self, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.labels_to_value('DONOTUSE') == 1

    @marvin_test_if(mark='skip', maps_release_only=dict(release=['MPL-4']))
    @pytest.mark.parametrize('names, expected',
                             [(['NOCOV', 'LOWCOV'], 3),
                              ('DONOTUSE', 1073741824)])
    def test_labels_to_value(self, maps_release_only, names, expected):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert ha.pixmask.labels_to_value(names) == expected

    @pytest.mark.parametrize('flag',
                             ['manga_target1',
                              'manga_target2',
                              'manga_target3',
                              'target_flags',
                              'pixmask'])
    def test_flag(self, flag, maps_release_only):
        ha = maps_release_only['emline_gflux_ha_6564']
        assert getattr(ha, flag, None) is not None


class TestEnhancedMap(object):

    def test_overridden_methods(self, galaxy):
        maps = Maps(plateifu=galaxy.plateifu)
        ha = maps['emline_gflux_ha_6564']
        nii = maps['emline_gflux_nii_6585']
        n2ha = nii / ha

        assert isinstance(n2ha, EnhancedMap)

        methods = ['_init_map_from_maps', '_get_from_file', '_get_from_db', '_get_from_api',
                   'inst_sigma_correction']

        for method in methods:
            with pytest.raises(AttributeError) as ee:
                meth = getattr(n2ha, method)
                meth()

            assert "'EnhancedMap' has no attribute '{}'.".format(method) in str(ee.value)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-02-19 23:37:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-20 16:28:28

from __future__ import print_function, division, absolute_import
import pytest
import os

from brain import bconfig
from brain.core.exceptions import BrainError, BrainUserWarning

# Tests on Brain Config

@pytest.fixture()
def netrc(monkeypatch, tmpdir):
    tmpnet = tmpdir.mkdir('netrc').join('.netrc')
    monkeypatch.setattr(bconfig, '_netrc_path', str(tmpnet))
    yield tmpnet


@pytest.fixture()
def goodnet(netrc):
    netrc.write('')
    os.chmod(bconfig._netrc_path, 0o600)
    yield netrc


@pytest.fixture()
def bestnet(goodnet):
    goodnet.write(write('data.sdss.org'))
    goodnet.write(write('api.sdss.org'))
    yield goodnet


def write(host):
    netstr = 'machine {0}\n'.format(host)
    netstr += '    login test\n'
    netstr += '    password test\n'
    netstr += '\n'
    return netstr


class TestVars(object):
    ''' test getting/setting variables '''

    @pytest.mark.parametrize('var, toval',
                             [('mode', 'remote'), ('access', 'collab'), ('compression', 'msgpack')])
    def test_set(self, monkeypatch, var, toval):
        defval = bconfig.__getattribute__(var)
        assert defval != toval
        monkeypatch.setattr(bconfig, var, toval)
        newval = bconfig.__getattribute__(var)
        assert newval == toval

    @pytest.mark.parametrize('var, toval',
                             [('mode', 'super'), ('access', 'always'), ('compression', 'donut')])
    def test_set_wrong(self, var, toval):
        with pytest.raises(ValueError) as cm:
            bconfig.__setattr__(var, toval)
        assert 'config.{0} must be'.format(var) in str(cm.value)

    def test_bad_access(self, netrc):
        assert bconfig.access == 'public'
        with pytest.raises(BrainError) as cm:
            bconfig.access = 'collab'
        assert 'No .netrc file found in your HOME directory!' in str(cm.value)
        assert bconfig.access == 'public'


class TestNetrc(object):
    ''' test the netrc access '''

    def test_no_netrc(self, netrc):
        with pytest.raises(BrainError) as cm:
            bconfig._check_netrc()
        assert 'No .netrc file found in your HOME directory!' in str(cm.value)

    def test_badperm_netrc(self, netrc):
        netrc.write('')
        with pytest.raises(BrainError) as cm:
            bconfig._check_netrc()
        assert 'your .netrc file does not have 600 permissions.' in str(cm.value)

    @pytest.mark.parametrize('host, msg',
                             [('data.sdss.org', 'api.sdss.org not found in netrc. You will not have remote access to SDSS data'),
                              ('api.sdss.org', 'data.sdss.org not found in netrc. You will not be able to download SDSS data')],
                             ids=['noapi', 'nodata'])
    def test_only_one_host(self, goodnet, host, msg):
        goodnet.write(write(host))
        with pytest.warns(BrainUserWarning) as cm:
            bconfig._check_netrc()

        assert msg in str(cm[0].message)

    def test_good_netrc(self, bestnet):
        assert bconfig._check_netrc() is True




#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, division, print_function

import os

import yaml

from tree.tree import Tree

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(name).
from .misc import log


NAME = 'tree'

# Loads config
with open(os.path.dirname(__file__) + '/etc/{0}.cfg'.format(NAME)) as ff:
    config = yaml.load(ff)


__version__ = '2.15.6dev'



# -*- coding: utf-8 -*-
#
# BMO documentation build configuration file, created by
# sphinx-quickstart on Fri May  5 01:30:21 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sphinx_bootstrap_theme
import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from tree import __version__
from pkg_resources import parse_version

# add a local path to the Sphinx search path
sys.path.insert(0, os.path.abspath('../../python/tree/misc/'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon', 'sphinx.ext.autosummary',
              'sphinx.ext.todo', 'sphinx.ext.viewcode', 'sphinx.ext.mathjax',
              'sphinx.ext.intersphinx', 'docutree']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
# source_suffix = '.rst'

source_parsers = {
    '.md': 'recommonmark.parser.CommonMarkParser',
}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'tree'
copyright = '{0}, {1}'.format('2017', 'Brian Cherinka')
author = 'Brian Cherinka'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The short X.Y version.
version = parse_version(__version__).base_version
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'py:obj'

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Intersphinx mappings
intersphinx_mapping = {'python': ('https://docs.python.org/3.6', None),
                       'astropy': ('http://docs.astropy.org/en/latest', None),
                       'numpy': ('http://docs.scipy.org/doc/numpy/', None)}

autodoc_mock_imports = ['_tkinter']
autodoc_member_order = 'groupwise'

napoleon_use_rtype = False
napoleon_use_ivar = True

rst_epilog = """
.. |numpy_array| replace:: Numpy array
.. |HDUList| replace:: :class:`~astropy.io.fits.HDUList`
"""


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'bootstrap'

html_sidebars = {}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # Navigation bar title. (Default: ``project`` value)
    'navbar_title': "SDSS: Tree",

    # Tab name for entire site. (Default: "Site")
    'navbar_site_name': "Site",

    # A list of tuples containing pages or urls to link to.
    # Valid tuples should be in the following forms:
    #    (name, page)                 # a link to a page
    #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
    #    (name, "http://example.com", True) # arbitrary absolute url
    # Note the "1" or "True" value above as the third argument to indicate
    # an arbitrary url.
    'navbar_links': [
    ],

    # Render the next and previous page links in navbar. (Default: true)
    'navbar_sidebarrel': False,

    # Render the current pages TOC in the navbar. (Default: true)
    'navbar_pagenav': False,

    # Tab name for the current pages TOC. (Default: "Page")
    'navbar_pagenav_name': "Page",

    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 2,

    # Include hidden TOCs in Site navbar?
    #
    # Note: If this is "false", you cannot have mixed ``:hidden:`` and
    # non-hidden ``toctree`` directives in the same page, or else the build
    # will break.
    #
    # Values: "true" (default) or "false"
    'globaltoc_includehidden': "true",

    # HTML navbar class (Default: "navbar") to attach to <div> element.
    # For black navbar, do "navbar navbar-inverse"
    'navbar_class': "navbar",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "true",

    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    'source_link_position': "",

    # Bootswatch (http://bootswatch.com/) theme.
    #
    # Options are nothing (default) or the name of a valid theme
    # such as "amelia" or "cosmo".
    'bootswatch_theme': "paper",

    # Choose Bootstrap version.
    # Values: "3" (default) or "2" (in quotes)
    'bootstrap_version': "3",
}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

html_favicon = './_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_sidebars = {'**': ['localtoc.html']}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = '{0}pdoc'.format('tree')


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, '{0}.tex'.format(project), u'{0} Documentation'.format(project),
     author, 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'tree', u'{0} Documentation'.format(project),
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, project, u'{0} Documentation'.format(project),
     author, project, 'One line description of project.',
     'Miscellaneous'),
]



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-14 14:31:36
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-14 19:03:32

from __future__ import print_function, division, absolute_import
import sys
import os
import argparse
import glob
import shutil


def check_sas_base_dir(root=None):
    ''' Check for the SAS_BASE_DIR environment variable

    Will set the SAS_BASE_DIR in your local environment
    or prompt you to define one if is undefined

    Parameters:
        root (str):
            Optional override of the SAS_BASE_DIR envvar

    '''
    sasbasedir = root or os.getenv("SAS_BASE_DIR")
    if not sasbasedir:
        sasbasedir = input('Enter a path for SAS_BASE_DIR: ')
    os.environ['SAS_BASE_DIR'] = sasbasedir


def write_header(term='bash', tree_dir=None, name=None):
    ''' Write proper file header in a given shell format

    Parameters:
        term (str):
            The type of shell header to write, can be "bash", "tsch", or "modules"
        tree_dir (str):
            The path to this repository
        name (str):
            The name of the configuration

    Returns:
        A string header to insert
    '''

    assert term in ['bash', 'tsch', 'modules'], 'term must be either bash, tsch, or module'

    product_dir = tree_dir.rstrip('/')
    base = 'export' if term == 'bash' else 'setenv'

    if term != 'modules':
        hdr = """# Set up tree/{0} for {1}
{2} TREE_DIR {3}
{2} TREE_VER {1}
{2} PATH $TREE_DIR/bin:$PATH
{2} PYTHONPATH $TREE_DIR/python:$PYTHONPATH
                """.format(name, term, base, product_dir)
    else:
        hdr = """#%Module1.0
proc ModulesHelp {{ }} {{
    global product version
    puts stderr "This module adds $product/$version to various paths"
}}
set name tree
set product tree
set version {1}
conflict $product
module-whatis "Sets up $product/$version in your environment"

set PRODUCT_DIR {0}
setenv [string toupper $product]_DIR $PRODUCT_DIR
setenv [string toupper $product]_VER $version
prepend-path PATH $PRODUCT_DIR/bin
prepend-path PYTHONPATH $PRODUCT_DIR/python

                """.format(product_dir, name)

    return hdr.strip()


def write_version(name):
    ''' Make the default modules version string '''
    modules_version = "#%Module1.0\nset ModulesVersion {0}".format(name)
    return modules_version


def write_file(environ, term='bash', out_dir=None, tree_dir=None):
    ''' Write a tree environment file

    Loops over the tree environ and writes them out to a bash, tsch, or
    modules file

    Parameters:
        environ (dict):
            The tree dictionary environment
        term (str):
            The type of shell header to write, can be "bash", "tsch", or "modules"
        tree_dir (str):
            The path to this repository
        out_dir (str):
            The output path to write the files (default is etc/)

    '''

    # get the proper name, header and file extension
    name = environ['default']['name']
    header = write_header(term=term, name=name, tree_dir=tree_dir)
    exts = {'bash': '.sh', 'tsch': '.csh', 'modules': '.module'}
    ext = exts[term]

    # shell command
    if term == 'bash':
        cmd = 'export {0}={1}\n'
    else:
        cmd = 'setenv {0} {1}\n'

    # write the environment config files
    filename = os.path.join(out_dir, name + ext)
    with open(filename, 'w') as f:
        f.write(header + '\n')
        for key, values in environ.items():
            if key != 'default':
                # write separator
                f.write('#\n# {0}\n#\n'.format(key))
                # write tree names and paths
                for tree_name, tree_path in values.items():
                    f.write(cmd.format(tree_name.upper(), tree_path))

    # write default .version file for modules
    modules_version = write_version(name)
    if term == 'modules' and environ['default']['current']:
        version_name = os.path.join(out_dir, '.version')
        with open(version_name, 'w') as f:
            f.write(modules_version)


def get_tree(config=None):
    ''' Get the tree for a given config

    Parameters:
        config (str):
            The name of the tree config to load

    Returns:
        a Python Tree instance
    '''
    path = os.path.dirname(os.path.abspath(__file__))
    pypath = os.path.realpath(os.path.join(path, '..', 'python'))
    if pypath not in sys.path:
        sys.path.append(pypath)
    os.chdir(pypath)
    from tree.tree import Tree
    tree = Tree(config=config)
    return tree


def copy_modules(filespath=None, modules_path=None, verbose=None):
    ''' Copy over the tree module files into your path '''

    # find or define a modules path
    if not modules_path:
        modulepath = os.getenv("MODULEPATH")
        if not modulepath:
            modules_path = input('Enter the root path for your module files:')
        else:
            split_mods = modulepath.split(':')
            if len(split_mods) > 1:
                if verbose:
                    print('Multiple module paths found.  Using top one: {0}'.format(split_mods[0]))
            modules_path = split_mods[0]

    # check for the tree module directory
    tree_mod = os.path.join(modules_path, 'tree')
    if not os.path.isdir(tree_mod):
        os.makedirs(tree_mod)

    # copy the modules into the tree
    if verbose:
        print('Copying modules from etc/ into {0}'.format(tree_mod))
    module_files = glob.glob(os.path.join(filespath, '*.module'))
    for mfile in module_files:
        base = os.path.splitext(os.path.basename(mfile))[0]
        tree_out = os.path.join(tree_mod, base)
        shutil.copy2(mfile, tree_out)

    # copy the default version into the tree
    version = os.path.join(filespath, '.version')
    if os.path.isfile(version):
        shutil.copy2(version, tree_mod)


def parse_args():
    ''' Parse the arguments '''

    parser = argparse.ArgumentParser(prog='setup_tree_modules', usage='%(prog)s [opts]')
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help='Print extra information.', default=False)
    parser.add_argument('-r', '--root', action='store', dest='root', default=os.getenv('SAS_BASE_DIR'),
                        help='Override the value of $SAS_BASE_DIR.', metavar='SAS_BASE_DIR')
    parser.add_argument('-t', '--treedir', action='store', dest='treedir', default=os.getenv('TREE_DIR'),
                        help='Override the value of $TREE_DIR.', metavar='TREE_DIR')
    parser.add_argument('-m', '--modulesdir', action='store', dest='modulesdir', default=os.getenv('MODULES_DIR'),
                        help='Your modules directory', metavar='MODULES_DIR')

    opts = parser.parse_args()

    return opts


def main(args):

    # parse arguments
    opts = parse_args()

    # get directories
    datadir = os.path.join(opts.treedir, 'data')
    etcdir = os.path.join(opts.treedir, 'etc')

    # config files
    configs = glob.glob(os.path.join(datadir, '*.cfg'))

    # check for the SAS_BASE_DIR
    check_sas_base_dir(root=opts.root)

    # Read and write the configuration files
    for cfgfile in configs:
        tree = get_tree(config=cfgfile)
        write_file(tree.environ, term='modules', out_dir=etcdir, tree_dir=opts.treedir)
        write_file(tree.environ, term='bash', out_dir=etcdir, tree_dir=opts.treedir)
        write_file(tree.environ, term='tsch', out_dir=etcdir, tree_dir=opts.treedir)

    # Setup the modules
    copy_modules(filespath=etcdir, verbose=opts.verbose)


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2018-07-20
# @Filename: test_quantities.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by:   andrews
# @Last modified time: 2018-10-19 14:10:15


import matplotlib
import numpy
import pytest
from astropy import units as u

from marvin.tests import marvin_test_if
from marvin.tools.quantities import DataCube, Spectrum


spaxel_unit = u.Unit('spaxel', represents=u.pixel, doc='A spectral pixel', parse_strict='silent')


@pytest.fixture(scope='function')
def datacube():
    """Produces a simple 3D array for datacube testing."""

    flux = numpy.tile([numpy.arange(1, 1001, dtype=numpy.float32)],
                      (100, 1)).T.reshape(1000, 10, 10)
    ivar = (1. / (flux / 100))**2
    mask = numpy.zeros(flux.shape, dtype=numpy.int)
    wave = numpy.arange(1, 1001)

    redcorr = numpy.ones(1000) * 1.5

    mask[50:100, 5, 5] = 2**10
    mask[500:600, 3, 3] = 2**4

    scale = 1e-3

    datacube = DataCube(flux, wave, ivar=ivar, mask=mask, redcorr=redcorr, scale=scale,
                        unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
                        pixmask_flag='MANGA_DRP3PIXMASK')

    yield datacube


@pytest.fixture(scope='function')
def spectrum():
    """Produces a simple 1D array for datacube testing."""

    flux = numpy.arange(1, 1001, dtype=numpy.float32)
    ivar = (1. / (flux / 100))**2
    mask = numpy.zeros(flux.shape, dtype=numpy.int)
    wave = numpy.arange(1, 1001)

    mask[50:100] = 2**10
    mask[500:600] = 2**4

    scale = 1e-3

    datacube = Spectrum(flux, wave, ivar=ivar, mask=mask, scale=scale,
                        unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
                        pixmask_flag='MANGA_DRP3PIXMASK')

    yield datacube


class TestDataCube(object):

    def test_datacube(self, datacube):

        assert datacube.value is not None
        assert datacube.ivar is not None
        assert datacube.mask is not None

        numpy.testing.assert_array_equal(datacube.value.shape, datacube.ivar.shape)
        numpy.testing.assert_array_equal(datacube.value.shape, datacube.mask.shape)

        assert datacube.pixmask is not None

    def test_masked(self, datacube):

        assert isinstance(datacube.masked, numpy.ma.MaskedArray)
        assert numpy.sum(datacube.masked.mask) == 50

        datacube.pixmask_flag = None
        assert numpy.sum(datacube.masked.mask) == 150

    def test_snr(self, datacube):

        assert datacube.snr[100, 5, 5] == pytest.approx(100)

    def test_error(self, datacube):

        numpy.testing.assert_almost_equal(datacube.error.value, numpy.sqrt(1 / datacube.ivar))
        assert datacube.error.unit == datacube.unit

        numpy.testing.assert_almost_equal(datacube.error.value, datacube.std.value)

    def test_descale(self, datacube):

        assert datacube.unit.scale == 1e-3

        descaled = datacube.descale()
        datacube.unit.scale == 1

        numpy.testing.assert_almost_equal(descaled.value, datacube.value * datacube.unit.scale)
        numpy.testing.assert_almost_equal(descaled.ivar, datacube.ivar / datacube.unit.scale**2)

    def test_redcorr(self, datacube):

        der = datacube.deredden()
        assert isinstance(der, DataCube)

        numpy.testing.assert_allclose(der.value, datacube.value * 1.5)
        numpy.testing.assert_allclose(der.ivar, datacube.ivar / 1.5**2)
        numpy.testing.assert_allclose(der.mask, datacube.mask)

        assert der.redcorr is None
        assert der.pixmask_flag == datacube.pixmask_flag

        new_redcorr = (numpy.ones(1000) * 2.)
        new_der = datacube.deredden(redcorr=new_redcorr)

        numpy.testing.assert_allclose(new_der.value, datacube.value * 2)
        numpy.testing.assert_allclose(new_der.ivar, datacube.ivar / 2**2)

        datacube.redcorr = None
        with pytest.raises(ValueError):
            datacube.deredden()

    def test_slice_datacube(self, datacube):

        new_datacube = datacube[:, 3:5, 3:5]

        assert isinstance(new_datacube, DataCube)
        numpy.testing.assert_almost_equal(new_datacube.value, datacube.value[:, 3:5, 3:5])
        numpy.testing.assert_almost_equal(new_datacube.ivar, datacube.ivar[:, 3:5, 3:5])
        numpy.testing.assert_almost_equal(new_datacube.mask, datacube.mask[:, 3:5, 3:5])
        numpy.testing.assert_almost_equal(new_datacube.redcorr, datacube.redcorr)
        assert new_datacube.pixmask_flag == datacube.pixmask_flag

    def test_slice_wave(self, datacube):

        new_datacube = datacube[10:100]

        assert isinstance(new_datacube, DataCube)
        numpy.testing.assert_almost_equal(new_datacube.value, datacube.value[10:100, :, :])
        numpy.testing.assert_almost_equal(new_datacube.ivar, datacube.ivar[10:100, :, :])
        numpy.testing.assert_almost_equal(new_datacube.mask, datacube.mask[10:100, :, :])
        numpy.testing.assert_almost_equal(new_datacube.redcorr, datacube.redcorr[10:100])
        assert new_datacube.pixmask_flag == datacube.pixmask_flag

    def test_slice_spectrum(self, datacube):

        new_spectrum = datacube[:, 5, 5]

        assert isinstance(new_spectrum, Spectrum)
        numpy.testing.assert_almost_equal(new_spectrum.value, datacube.value[:, 5, 5])
        numpy.testing.assert_almost_equal(new_spectrum.ivar, datacube.ivar[:, 5, 5])
        numpy.testing.assert_almost_equal(new_spectrum.mask, datacube.mask[:, 5, 5])
        assert new_spectrum.pixmask_flag == datacube.pixmask_flag

    @marvin_test_if(mark='include', cube={'plateifu': '8485-1901',
                                          'data_origin': 'file',
                                          'initial_mode': 'local'})
    def test_cube_quantities(self, cube):

        assert cube.flux is not None

        assert isinstance(cube.flux, numpy.ndarray)
        assert isinstance(cube.flux, DataCube)

        assert isinstance(cube.spectral_resolution, Spectrum)

        if cube.release in ['MPL-4', 'MPL-5']:
            with pytest.raises(AssertionError) as ee:
                cube.spectral_resolution_prepixel
            assert 'spectral_resolution_prepixel is not present in his MPL version' in str(ee)
        else:
            assert isinstance(cube.spectral_resolution_prepixel, Spectrum)

        assert cube.flux.pixmask.values_to_bits(3) == [0, 1]

        assert cube.flux.pixmask.values_to_labels(3) == ['NOCOV', 'LOWCOV']

    @pytest.mark.parametrize('names, expected', [(['NOCOV', 'LOWCOV'], 3),
                                                 ('DONOTUSE', 1024)])
    def test_labels_to_value(self, cube, names, expected):
        assert cube.flux.pixmask.labels_to_value(names) == expected

    @marvin_test_if(mark='include', modelcube={'plateifu': '8485-1901',
                                               'data_origin': 'file',
                                               'initial_mode': 'local'})
    def test_modelcube_quantities(self, modelcube):

        for mc in modelcube.datamodel:
            if hasattr(modelcube, mc.name):
                modelcube_quantity = getattr(modelcube, mc.name)
                assert isinstance(modelcube_quantity, DataCube)
                assert modelcube_quantity.pixmask_flag == 'MANGA_DAPSPECMASK'


class TestSpectrum(object):

    def test_spectrum(self, spectrum):

        assert spectrum.value is not None
        assert spectrum.ivar is not None
        assert spectrum.mask is not None

        numpy.testing.assert_array_equal(spectrum.value.shape, spectrum.ivar.shape)
        numpy.testing.assert_array_equal(spectrum.value.shape, spectrum.mask.shape)

        assert spectrum.pixmask is not None

    def test_masked(self, spectrum):

        assert isinstance(spectrum.masked, numpy.ma.MaskedArray)
        assert numpy.sum(spectrum.masked.mask) == 50

        spectrum.pixmask_flag = None
        assert numpy.sum(spectrum.masked.mask) == 150

    def test_snr(self, spectrum):

        assert spectrum.snr[100] == pytest.approx(100)

    def test_error(self, spectrum):

        numpy.testing.assert_almost_equal(spectrum.error.value, numpy.sqrt(1 / spectrum.ivar))
        assert spectrum.error.unit == spectrum.unit

        numpy.testing.assert_almost_equal(spectrum.error.value, spectrum.std.value)

    def test_descale(self, spectrum):

        assert spectrum.unit.scale == 1e-3

        descaled = spectrum.descale()
        spectrum.unit.scale == 1

        numpy.testing.assert_almost_equal(descaled.value, spectrum.value * spectrum.unit.scale)
        numpy.testing.assert_almost_equal(descaled.ivar, spectrum.ivar / spectrum.unit.scale**2)

    def test_slice_spectrum(self, spectrum):

        new_spectrum = spectrum[10:100]

        assert isinstance(new_spectrum, Spectrum)
        numpy.testing.assert_almost_equal(new_spectrum.value, spectrum.value[10:100])
        numpy.testing.assert_almost_equal(new_spectrum.ivar, spectrum.ivar[10:100])
        numpy.testing.assert_almost_equal(new_spectrum.mask, spectrum.mask[10:100])
        assert new_spectrum.pixmask_flag == spectrum.pixmask_flag

    @marvin_test_if(mark='include', cube={'plateifu': '8485-1901',
                                          'data_origin': 'file',
                                          'initial_mode': 'local'})
    def test_cube_quantities(self, cube):

        for sp in cube.datamodel.spectra:
            cube_quantity = getattr(cube, sp.name)
            assert isinstance(cube_quantity, Spectrum)
            assert cube_quantity.pixmask_flag is None

    def test_plot(self, spectrum):

        ax = spectrum.plot(show_std=True)
        assert isinstance(ax, matplotlib.axes.Axes)

    def test_plot_no_std_no_mask(self):
        sp = Spectrum(numpy.random.randn(1000), wavelength=numpy.arange(1000))
        sp.plot()

    def test_plot_no_std(self):

        mask = numpy.zeros(1000, dtype=numpy.int)
        mask[50:100] = 2**10
        mask[500:600] = 2**4

        sp = Spectrum(
            flux=numpy.random.randn(1000),
            wavelength=numpy.arange(1000),
            mask=mask,
            pixmask_flag='MANGA_DRP3PIXMASK',
        )
        sp.plot()

    def test_plot_no_mask(self):
        flux = numpy.random.randn(1000)
        ivar = (1. / (flux / 100))**2

        sp = Spectrum(
            flux=flux,
            wavelength=numpy.arange(1000),
            ivar=ivar,
        )
        sp.plot()




from functools import wraps
from brain.core.exceptions import BrainError

try:
    from sdss_access.path import Path
except ImportError:
    Path = None

# General Decorators
__all__ = ['public', 'parseRoutePath', 'checkPath']


def public(f):
    ''' Decorator to declare a route public '''
    f.is_public = True
    return f


def parseRoutePath(f):
    ''' Decorator to parse generic route path '''
    @wraps(f)
    def decorated_function(inst, *args, **kwargs):
        if 'path' in kwargs and kwargs['path']:
            for kw in kwargs['path'].split('/'):
                if len(kw) == 0:
                    continue
                var, value = kw.split('=')
                kwargs[var] = value
        kwargs.pop('path')
        return f(inst, *args, **kwargs)
    return decorated_function


def checkPath(func):
    '''Decorator that checks if sdss_access Path has been imported '''

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not Path:
            raise BrainError('sdss_access is not installed')
        else:
            return func(*args, **kwargs)
    return wrapper


def check_auth(func):
    ''' Decorator that checks if a valid netrc file exists

    Function Decorator to check if a valid netrc file exists.
    If not it raises an error.  Otherwise it
    returns the function and proceeds as normal.

    Returns:
        The decorated function

    Raises:
        BrainError: You are not authorized to access the SDSS collaboration

    Example:
        >>>
        >>> @check_auth
        >>> def my_function():
        >>>     return 'I am working function'
        >>>

    '''

    from brain import bconfig
    @wraps(func)
    def wrapper(*args, **kwargs):
        # iscollab = bconfig.access == 'collab'
        valid_netrc = bconfig._check_netrc()
        if valid_netrc:
            return func(*args, **kwargs)
        else:
            raise BrainError('You are not authorized to access the SDSS collaboration')

    return wrapper




from __future__ import print_function
import requests
import os
from requests.auth import AuthBase, _basic_auth_str
from requests.utils import get_netrc_auth
from brain.core.exceptions import BrainError, BrainApiAuthError, BrainNotImplemented
from brain import bconfig
from brain.utils.general import uncompress_data
from brain.core.core import URLMapDict
try:
    from urlparse import urlsplit, urlunsplit
except ImportError:
    from urllib.parse import urlsplit, urlunsplit
try:
    from cachecontrol import CacheControl as cache
except ImportError:
    cache = None

configkeys = []


class BrainInteraction(object):
    """ This class defines convenience wrappers for the Brain RESTful API """

    def __init__(self, route, params=None, request_type='post', auth='token',
                 timeout=(3.05, 300), headers=None, stream=None,
                 datastream=None, send=True, base=None):
        self.results = None
        self.response_time = None
        self.route = route
        self.params = params
        self.request_type = request_type
        self.timeout = timeout
        self.stream = stream
        self.datastream = datastream
        self.headers = headers if headers is not None else {}
        self.statuscodes = {200: 'Ok', 401: 'Authentication Required', 404: 'URL Not Found',
                            500: 'Internal Server Error', 405: 'Method Not Allowed',
                            400: 'Bad Request', 502: 'Bad Gateway', 504: 'Gateway Timeout',
                            422: 'Unprocessable Entity', 429: 'Rate Limit Exceeded'}
        self.compression = self.params['compression'] if self.params and \
            'compression' in self.params else bconfig.compression

        base = base if base else bconfig.sasurl
        self.url = urljoin(base, route) if self.route else None

        # set request Session
        self._setRequestSession()

        # set authentication
        self.setAuth(authtype=auth)

        # sends the request
        if self.url and send:
            self._sendRequest(request_type)
        elif not self.url and send:
            raise BrainError('No route and/or url specified {0}'.format(self.url))

    def __repr__(self):
        return ('Interaction(route={0}, params={1}, request_type={2})'
                .format(self.route, repr(self.params), self.request_type))

    def _setRequestSession(self):
        ''' creates or sets the Brain requests Session object '''
        if isinstance(bconfig.request_session, type(None)):
            if not cache:
                bconfig.request_session = requests.Session()
            else:
                bconfig.request_session = cache(requests.Session())
            self.session = bconfig.request_session
        else:
            self.session = bconfig.request_session

    def _closeRequestSession(self):
        ''' closes the Brain requests Session object'''
        self.session.close()
        bconfig.request_session.close()
        bconfig.request_session = None

    def setAuth(self, authtype='netrc'):
        ''' set the session authentication '''
        self.authtype = authtype
        if authtype:
            self.session.auth = BrainAuth(self.authtype)

    def _decode_stream(self, content):
        ''' Decode the content string for a data stream

        Uncompresses the response content either using JSON or msgpack.
        Will uncompress the entire response content in single go. If
        datastream is True, then the content is streamed back using a
        generator and has been compressed row-by-row. See the generator
        query fxn in marvin/api/query.py.

        Parameters:
            content (str):
                The response content string

        Returns:
            The uncompressed content data

        '''

        if self.datastream:
            # decoding a generator content stream
            # since content is a single string, must split on the row separator
            data = [uncompress_data(row, uncompress_with=self.compression) for row in content.split(';\n') if row]
            # data is expected to be in a dictionary key called 'data'
            out = {}
            out['data'] = data
        else:
            # decoding a normal content stream
            out = uncompress_data(content, uncompress_with=self.compression)
        return out

    def _get_json(self, response):
        ''' Try to extract any json data

        Tries to use the built in json fxn to extract
        the response content

        Parameters:
            response:
                The full response

        Returns:
            The JSON data
        '''
        try:
            json = response.json()
        except ValueError as e:
            json = None
        return json

    def _get_data(self, response, chunksize=None, dtype=None):
        ''' Get json/binary data from the response content

        Tries to extract the data from the response content.
        If stream is set, then streams the response content in chunks and extracts
        the data, to save on client memory.  Otherwise reads the entire response
        content into memory at once.

        Parameters:
            response:
                The full response
            chunksize (int):
                The unit of the chunk size in bytes.  When None, chunksize determined by
                interal magic. Default is None
            dtype (str):
                The data type.  Either json or other (e.g. binary content using msgpack).

        Returns:
            The uncompressed data as a list.

        '''

        if self.stream:
            # retrieves response data in chunks to minimize client memory
            resstring = ''.join([bytes.decode(chunk) for chunk in response.iter_content(chunk_size=chunksize)])
            data = self._decode_stream(resstring)
        else:
            # retrieves response data all at once
            if dtype == 'json':
                data = self._get_json(response)
            else:
                data = uncompress_data(response.content, uncompress_with='msgpack')
        return data

    def _get_content(self, response):
        ''' Get the response content

        Gets the response content either as a JSON
        or binary data from msgpack

        Parameters:
            response:
                The full response

        Returns:
            The uncompressed data

        '''

        content_type = response.headers['Content-Type']
        if 'json' in content_type:
            data = self._get_data(response, dtype='json')
        elif 'octet-stream' in content_type:
            data = self._get_data(response)
        else:
            data = response.content

        return data

    def _checkResponse(self, response):
        ''' Checks the response for proper http status code '''

        # check for bad status
        try:
            isbad = response.raise_for_status()
        except requests.HTTPError as http:
            self.status_code = response.status_code
            errmsg = 'Error accessing {0}: {1}-{2}'.format(response.url, response.status_code,
                                                           self.statuscodes[response.status_code])
            self.results = {'http_status_code': response.status_code, 'message': errmsg}
            json_data = self._get_content(response)

            if self.status_code == 401:
                if self.authtype == 'netrc':
                    msg = 'Please create or check credentials in your local .netrc file'
                elif self.authtype == 'oauth':
                    msg = 'Please check your Oauth authentication'
                elif self.authtype == 'http':
                    msg = 'Please check your http/digest authentication'
                elif self.authtype == 'token':
                    err = json_data['msg'] if b'msg' in json_data else str(json_data)
                    msg = '{0}. Please check your token or login again for a fresh one.'.format(err)
                else:
                    msg = 'Please check your authentication method.'
                errmsg = json_data['error'] if b'error' in json_data else ''
                self._closeRequestSession()
                raise BrainApiAuthError('API Authentication Error: {0}. {1}'.format(msg, errmsg))
            elif self.status_code == 422:
                self._closeRequestSession()
                raise BrainError('Requests Http Status Error: {0}\nValidation Errors:\n{1}'.format(http, json_data))
            else:
                self._closeRequestSession()
                if b'api_error' in json_data:
                    apijson = json_data['api_error']
                    errmsg = '{0}\n{1}'.format(apijson['message'], apijson['traceback']) if 'message' in apijson else '{0}'.format(apijson['traceback'])
                elif b'error' in json_data:
                    err = json_data['error']
                    errmsg = '{0}'.format(err)
                raise BrainError('Requests Http Status Error: {0}\n{1}'.format(http, errmsg))
        else:
            # Not bad
            assert isbad is None, 'Http status code should not be bad'
            assert response.ok is True, 'Ok status should be true'

            # if 'route' not in self.url and 'clean' not in self.url:
            #     self._closeRequestSession()
            #     raise BrainError('test error is now raised here')

            self.status_code = response.status_code
            self.results = self._get_content(response)
            self.response_time = response.elapsed
            if not self.results:
                self.results = response.text
                self._closeRequestSession()
                raise BrainError('Response not in JSON format. {0}'.format(self.results))

            # Raises an error if status is -1
            if 'status' in self.results and self.results['status'] == -1:
                errorMsg = 'no error message provided' \
                    if 'error' not in self.results else self.results['error']
                self._check_for_traceback()
                self._closeRequestSession()
                raise BrainError('Something went wrong on the server side: {0}'.format(errorMsg))

    def _check_for_traceback(self):
        ''' checks the response for a traceback in the results response '''
        bconfig.traceback = self.results.get('traceback', None)

    def _sendRequest(self, request_type):
        ''' sends the api requests to the server '''
        assert request_type in ['get', 'post'], 'Valid request types are "get" and "post".'

        # Loads the local config parameters
        self._loadConfigParams()

        # Send the request
        try:
            if request_type == 'get':
                self._response = self.session.get(self.url, params=self.params, timeout=self.timeout,
                                                  headers=self.headers, stream=self.stream)
            elif request_type == 'post':
                self._response = self.session.post(self.url, data=self.params, timeout=self.timeout,
                                                   headers=self.headers, stream=self.stream)
        except requests.Timeout as rt:
            self._closeRequestSession()
            errmsg = 'Your request took longer than 5 minutes and timed out. Please try again or simplify your request.'
            raise BrainError('Requests Timeout Error: {0}\n{1}'.format(rt, errmsg))
        except requests.URLRequired as urlreq:
            self._closeRequestSession()
            raise BrainError('Requests Valid URL Required: {0}'.format(urlreq))
        except requests.ConnectionError as con:
            self._closeRequestSession()
            raise BrainError('Requests Connection Error: {0}'.format(con))
        except requests.RequestException as req:
            self._closeRequestSession()
            raise BrainError('Ambiguous Requests Error: {0}'.format(req))
        else:
            # Check the response if it's good
            self._checkResponse(self._response)

    def getData(self, astype=None):
        data = self.results['data'] if 'data' in self.results else None

        if astype and data:
            try:
                return astype(data)
            except Exception as e:
                raise Exception('Failed: {0}, {1}'.format(e, data))
        else:
            return data

    # This should probably get moved into the server side code
    def _preloadResults(self):
        for key in configkeys:
            self.results[key] = bconfig.__getattribute__(key)

    def checkMyConfig(self):
        return {k: self.results[k] if k in self.results else '' for k in configkeys}

    def _loadConfigParams(self):
        """Load the local configuration into a parameters dictionary to be sent with the request"""
        self.params.update({'session_id': bconfig.session_id})

    def getRouteMap(self):
        """Retrieve the URL routing map if it exists."""

        if 'urlmap' in self.results:
            return URLMapDict(self.results['urlmap'])
        else:
            return URLMapDict()


class BrainAuth(AuthBase):
    ''' This is a custom requests authorization class

    This class is used to ensure that token authentication takes
    precedence over the netrc file in the request.  See
    https://github.com/requests/requests/issues/3929 and
    `Request Custom Headers <http://docs.python-requests.org/en/master/user/quickstart/#custom-headers>_`

    Parameters:
        authtype (str):
            the authentication method used for the API.  Currently set as default to use
            JSON Web token authentication.

    '''

    def __init__(self, authtype, **kwargs):
        self.authtype = authtype

    def __call__(self, r):
        ''' checks the authentication when the auth is called '''

        # check access and authentication
        if self.authtype == 'token':
            assert bconfig.token is not None, 'You must have a valid token set to use the API.  Please login.'
            r.headers['Authorization'] = 'Bearer {0}'.format(bconfig.token)
        elif self.authtype == 'http':
            from requests_toolbelt import GuessAuth
            r.auth = GuessAuth('user', 'passwd')
        elif self.authtype == 'netrc' and bconfig.has_netrc:
            netauth = get_netrc_auth(r.url)
            if netauth:
                r.headers['Authorization'] = _basic_auth_str(*get_netrc_auth(r.url))
        elif self.authtype == 'oauth':
            raise BrainNotImplemented('OAuth authentication')
        return r


def urljoin(url1, url2):
    ''' custom function to join two url paths '''

    e = urlsplit(url1)
    t = urlsplit(url2)
    final = urlunsplit(tuple(strjoin(*z) for z in zip(e, t)))
    return final


def strjoin(str1, str2):
    ''' joins two url strings '''
    if not str2.startswith(str1):
        f = os.path.join(str1, str2.lstrip('/')) if str2 else str1
    else:
        f = str2
    return f



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-02-12 18:00:11
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-20 18:20:59

from __future__ import print_function, division, absolute_import
from marvin.api.api import Interaction
import pytest


auths = [None, 'token', 'netrc']


@pytest.fixture(params=auths)
def mint(request):
    base = 'https://lore.sdss.utah.edu/'
    url = '/marvin/api/general/getroutemap/'
    if request.param is None:
        pytest.skip("no auth should fail")
    ii = Interaction(url, auth=request.param, send=False, base=base)
    yield ii
    ii = None


class TestInteraction(object):

    def test_auth(self, mint):
        assert mint.authtype in auths
        if mint.authtype:
            assert mint.authtype == mint.session.auth.authtype

    def test_auth_fail(self):
        base = 'https://lore.sdss.utah.edu/'
        url = '/marvin/api/general/getroutemap/'
        with pytest.raises(AssertionError) as cm:
            ii = Interaction(url, auth=None, send=False, base=base)
        errmsg = 'Must have an authorization type set for collab access to MPLs!'
        assert errmsg in str(cm.value)





# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-09-27 11:08:07
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-14 19:09:17

from __future__ import print_function, division, absolute_import
import os
from invoke import Collection, task


# This file contains tasks that can be easily run from the shell terminal using the Invoke python package
# If you do not have invoke, install it with pip install
# To list the tasks available, type invoke --list from the top-level repo directory

@task
def clean_docs(ctx):
    ''' Cleans up the Sphinx docs '''
    print('Cleaning the docs')
    ctx.run("rm -rf docs/sphinx/_build")


@task(clean_docs)
def build_docs(ctx):
    ''' Builds the Sphinx docs '''
    print('Building the docs')
    os.chdir('docs/sphinx')
    ctx.run("make html")


@task(build_docs)
def show_docs(ctx):
    """Shows the Sphinx docs"""
    print('Showing the docs')
    os.chdir('_build/html')
    ctx.run('open ./index.html')


@task
def clean(ctx):
    ''' Cleans up the crap before a Pip build '''
    print('Cleaning')
    ctx.run("rm -rf htmlcov")
    ctx.run("rm -rf build")
    ctx.run("rm -rf dist")


@task(clean)
def deploy(ctx):
    ''' Deploy the project to pypi '''
    print('Deploying to Pypi!')
    ctx.run("python setup.py sdist bdist_wheel --universal")
    ctx.run("twine upload dist/*")


@task
def setup_tree(ctx, verbose=None, root=None, tree_dir=None, modules_dir=None):
    ''' Sets up the SDSS tree enviroment '''
    print('Setting up the tree')
    ctx.run('python bin/setup_tree.py -t {0} -r {1} -m {2}'.format(tree_dir, root, modules_dir))


# create a collection of tasks
ns = Collection(clean, deploy, setup_tree)

# create a sub-collection for the doc tasks
docs = Collection('docs')
docs.add_task(build_docs, 'build')
docs.add_task(clean_docs, 'clean')
docs.add_task(show_docs, 'show')
ns.add_collection(docs)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-08-01 17:50:48
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-08-08 14:01:33

from __future__ import print_function, division, absolute_import

import numpy as np
import pytest
from marvin.tools.image import Image


IMCOORDS = np.array([[275.38201798, 275.38201798],
                     [275.38201798, 281.0],
                     [275.38201798, 286.61798202],
                     [281.0, 275.38201798],
                     [281.0, 281.0],
                     [281.0, 286.61798202],
                     [286.61798202, 275.38201798],
                     [286.61798202, 281.0],
                     [286.61798202, 286.61798202]])

@pytest.fixture(scope='session')
def image():
    im = Image('8485-1901')
    yield im
    im = None


class TestImage(object):

    def test_getimage(self, image):
        assert isinstance(image, Image)
        assert image.data is not None
        assert image.plateifu == '8485-1901'

    @pytest.mark.parametrize('plateifu, origin, dir3d',
                             [('8485-1901', 'file', 'stack'),
                              ('8116-1901', 'api', 'mastar')])
    def test_origins(self, plateifu, origin, dir3d):
        im = Image(plateifu)
        assert im.plateifu == plateifu
        assert im.data_origin == origin
        assert dir3d in im._getFullPath()

    def test_release(self, plateifu):
        im = Image(plateifu, release='MPL-5')
        assert im.release == 'MPL-5'

    @pytest.mark.parametrize('plateifu, release',
                             [('8485-1901', 'MPL-6'),
                              ('8116-1901', 'MPL-6'),
                              ('8116-1901', 'MPL-4')])
    def test_attributes(self, plateifu, release):
        im = Image(plateifu, release=release)
        if release == 'MPL-4':
            assert im.wcs is None
            assert im.header is None
            assert im.ra is None
            assert not hasattr(im, 'bundle')
        else:
            assert im.wcs is not None
            assert im.header is not None
            assert im.ra is not None
            assert hasattr(im, 'bundle')

    @pytest.mark.parametrize('cubecoord, imcoord',
                             [((17, 17), (281, 281)),
                              ((18, 18), (286.61798202, 286.61798202)),
                              ((9.99826865, 6.34193863), (241.66439915, 221.12320284))],
                             ids=['center', 'off', 'hexedge'])
    def test_wcs(self, image, cubecoord, imcoord):
        cube = image.getCube()
        wcs = cube.wcs.celestial
        cube_radec = wcs.all_pix2world([cubecoord], 0)
        im_radec = image.wcs.all_pix2world([imcoord], 1)
        assert im_radec == pytest.approx(cube_radec, rel=1e-6)

    def test_wcs_aperture(self, image):
        cube = image.getCube()
        wcs = cube.wcs.celestial
        aper = cube.getAperture((17, 17), 1)
        coords = zip(*np.where(aper.mask))
        im_radec = image.wcs.all_pix2world(IMCOORDS, 1)
        cube_radec = wcs.all_pix2world(coords, 0)
        assert im_radec == pytest.approx(cube_radec, rel=1e-6)

    def test_saveimage(self, image, temp_scratch):
        file = temp_scratch.join('test_image.png')
        image.save(str(file))
        assert file.check() is True

    def test_new_cutout(self, image):
        wcs = image.wcs
        image.get_new_cutout(100, 100, scale=0.192)
        assert wcs != image.wcs

    def test_fromlist(self):
        plateifu = ['8485-1901', '7443-12701']
        images = Image.from_list(plateifu)
        names = [im.plateifu for im in images]
        assert plateifu == names

    def test_getrandom(self):
        images = Image.get_random(2)
        assert len(images) == 2

    def test_byplate(self):
        images = Image.by_plate(8485)
        assert isinstance(images, list)
        assert images[0].plate == 8485


class TestBundle(object):

    def test_skies(self, image):
        assert image.bundle.skies is None
        image.bundle.get_sky_coordinates()
        assert image.bundle.skies is not None
        assert len(image.bundle.skies) == 2









#!/usr/bin/env python
# encoding: utf-8

'''
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-17 17:46:57
    Last Modified On: 2016-02-17 17:46:57 by Brian

'''
from __future__ import print_function
from __future__ import division
from brain.api.base import BrainBaseView
from brain.utils.general import build_routemap
from marvin import config
from marvin.api import ArgValidator, set_api_decorators
from flask import current_app


arg_validate = ArgValidator(urlmap=None)


class BaseView(BrainBaseView):
    '''Super Class for all API Views to handle all global API items of interest'''
    decorators = set_api_decorators()

    def add_config(self):
        utahconfig = {'utahconfig': {'mode': config.mode, 'release': config.release}}
        self.update_results(utahconfig)

    def _pop_args(self, kwargs, arglist=None):
        if arglist:
            arglist = [arglist] if not isinstance(arglist, (list, tuple)) else arglist

            for item in arglist:
                tmp = kwargs.pop(item, None)

        return kwargs

    def before_request(self, *args, **kwargs):

        # try to get a local version of the urlmap for the arg_validator
        if not arg_validate.urlmap:
            urlmap = build_routemap(current_app)
            config.urlmap = urlmap
            arg_validate.urlmap = urlmap

        super(BaseView, self).before_request(*args, **kwargs)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-03-08 18:08:34
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-20 18:12:02

from __future__ import print_function, division, absolute_import
import pytest
import os
import warnings

try:
    from urlparse import urlsplit, urlunsplit
except ImportError:
    from urllib.parse import urlsplit, urlunsplit

from marvin import config
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from brain import bconfig
from brain.core.exceptions import BrainError, BrainUserWarning


@pytest.fixture()
def setapi(monkeypatch):
    monkeypatch.setitem(bconfig._custom_config, 'use_test', False)
    bconfig._set_api_urls()


@pytest.fixture()
def netrc(monkeypatch, tmpdir):
    config.access = 'public'
    tmpnet = tmpdir.mkdir('netrc').join('.netrc')
    monkeypatch.setattr(bconfig, '_netrc_path', str(tmpnet))
    yield tmpnet


@pytest.fixture()
def goodnet(netrc):
    netrc.write('')
    os.chmod(bconfig._netrc_path, 0o600)
    yield netrc


@pytest.fixture()
def bestnet(goodnet):
    goodnet.write(write('data.sdss.org'))
    goodnet.write(write('api.sdss.org'))
    config._check_access()
    config.access = 'collab'
    config.setRelease('MPL-6')
    yield goodnet


def write(host):
    netstr = 'machine {0}\n'.format(host)
    netstr += '    login test\n'
    netstr += '    password test\n'
    netstr += '\n'
    return netstr


@pytest.fixture()
def initconfig(monkeypatch):
    monkeypatch.delattr(config, '_tree')
    monkeypatch.setattr(config, '_release', None)


@pytest.fixture()
def set_default(monkeypatch, request):
    monkeypatch.setattr(config, '_release', request.param)
    monkeypatch.setitem(config._custom_config, 'default_release', request.param)
    config._check_config()


class TestVars(object):
    ''' test getting/setting variables '''

    @pytest.mark.parametrize('var, toval',
                             [('mode', 'remote'), ('access', 'public')])
    def test_set(self, monkeypatch, var, toval):
        defval = config.__getattribute__(var)
        assert defval != toval
        monkeypatch.setattr(config, var, toval)
        newval = config.__getattribute__(var)
        assert newval == toval

    @pytest.mark.parametrize('var, toval',
                             [('mode', 'super'), ('access', 'always')])
    def test_set_wrong(self, var, toval):
        with pytest.raises(ValueError) as cm:
            config.__setattr__(var, toval)
        assert 'config.{0} must be'.format(var) in str(cm.value)


class TestAccess(object):

    def test_bad_access(self, netrc):
        assert config.access == 'public'
        with pytest.raises(BrainError) as cm:
            config.access = 'collab'
        assert 'No .netrc file found in your HOME directory!' in str(cm.value)
        assert config.access == 'public'

    def test_public_access(self, bestnet):
        assert config.access == 'collab'
        assert 'MPL-5' in config._allowed_releases
        assert 'DR15' in config._allowed_releases
        config.access = 'public'
        assert 'MPL-5' not in config._allowed_releases
        assert 'DR15' in config._allowed_releases

    def test_tree(self, bestnet):
        assert config.access == 'collab'
        assert 'mangawork' in os.environ['MANGA_SPECTRO_REDUX']
        assert 'MPL' in config.release

        config.access = 'public'
        assert 'sas/dr' in os.environ['MANGA_SPECTRO_REDUX']
        assert 'DR' in config.release

        config.access = 'collab'
        assert 'sas/dr' in os.environ['MANGA_SPECTRO_REDUX']
        assert 'DR' in config.release


class TestReleases(object):

    @pytest.mark.parametrize('release', [('dr15'), ('mpl-5')])
    def test_tree(self, bestnet, release):
        assert config.access == 'collab'
        assert 'mangawork' in os.environ['MANGA_SPECTRO_REDUX']
        assert 'MPL' in config.release

        config.setRelease(release)
        if 'mpl' in release:
            assert 'mangawork' in os.environ['MANGA_SPECTRO_REDUX']
        else:
            assert release in os.environ['MANGA_SPECTRO_REDUX']

        assert config.release == release.upper()

    @pytest.mark.parametrize('release', [('dr15'), ('mpl-6')])
    def test_drpall(self, bestnet, release):
        assert 'mangawork' in config.drpall
        config.setRelease(release)
        if config.drpall:
            word = 'mangawork' if 'mpl' in release else release
            assert word in config.drpall

    def test_invalid_release(self):
        with pytest.raises(MarvinError) as cm:
            config.setRelease('MPL-1')
        assert 'trying to set an invalid release version.' in str(cm.value)

    def test_invalid_dr(self):
        with pytest.raises(AssertionError) as cm:
            config.setDR('MPL-5')
        assert 'Must specify a DRXX version!' in str(cm.value)

    def test_invalid_mpl(self):
        with pytest.raises(AssertionError) as cm:
            config.setMPL('DR15')
        assert 'Must specify an MPL-X version!' in str(cm.value)


class TestNetrc(object):
    ''' test the netrc access '''

    @pytest.mark.parametrize('host, msg',
                             [('data.sdss.org', 'api.sdss.org not found in netrc. You will not have remote access to SDSS data'),
                              ('api.sdss.org', 'data.sdss.org not found in netrc. You will not be able to download SDSS data')],
                             ids=['noapi', 'nodata'])
    def test_only_one_host(self, goodnet, host, msg):
        goodnet.write(write(host))
        with pytest.warns(BrainUserWarning) as cm:
            config._check_access()

        assert msg in str(cm[0].message)

    def test_good_netrc(self, bestnet):
        config._check_access()
        assert config.access == 'collab'


class TestConfig(object):

    def test_exists(self):
        assert config is not None

    def test_bad_login(self):
        config.access = 'public'
        with pytest.raises(AssertionError) as cm:
            config.login()
        assert 'You must have collaboration access to login.' in str(cm.value)

    @pytest.mark.parametrize('defrel, exprel',
                             [('DR20', 'MPL-7'), ('bad_release', 'MPL-7')])
    def test_bad_default_release(self, initconfig, defrel, exprel):
        ''' this tests some initial conditions on config '''
        config._release = defrel
        config._check_access()
        msg = 'Release {0} is not in the allowed releases.  Switching to {1}'.format(defrel, exprel)
        with pytest.warns(MarvinUserWarning):
            warnings.warn(msg, MarvinUserWarning)
        assert config.release == exprel


@pytest.mark.usefixtures('setapi')
class TestSasUrl(object):

    def test_sasurl_nonetrc(self, initconfig, netrc):
        assert 'DR' in config.release
        assert 'dr15.sdss.org/' in config.sasurl

    @pytest.mark.parametrize('release',
                             [('MPL-6'), ('DR15')],
                             ids=['collab', 'public'])
    def test_sasurl(self, bestnet, release):
        assert 'sas.sdss.org' in config.sasurl
        config.setRelease(release)
        sasurl = 'dr15.sdss.org' if 'DR' in release else 'sas.sdss.org'
        assert sasurl in config.sasurl

    @pytest.mark.parametrize('sas, exp',
                             [('utah', 'sas.sdss.org'),
                              ('public', 'dr15.sdss.org'),
                              ('test', 'lore.sdss.utah.edu'),
                              ('testpub', 'lore.sdss.utah.edu/public'),
                              ('local', 'localhost')],
                             )
    def test_sasurl_switch(self, sas, exp):
        public = sas == 'public'
        test = sas == 'test'
        if sas == 'testpub':
            public = test = True
        sas = 'utah' if sas != 'local' else sas
        config.switchSasUrl(sas, public=public, test=test)
        assert exp in config.sasurl

    @pytest.mark.parametrize('sas, exp',
                             [('utah', 'https://sas.sdss.org/marvin/api/cubes/8485-1901/'),
                              ('public', 'https://dr15.sdss.org/marvin/api/cubes/8485-1901/'),
                              ('test', 'https://lore.sdss.utah.edu/marvin/api/cubes/8485-1901/'),
                              ('testpub', 'https://lore.sdss.utah.edu/public/marvin/api/cubes/8485-1901/'),
                              ('local', 'http://localhost:5000/marvin/api/cubes/8485-1901/')],
                             )
    def test_sasurl_join(self, sas, exp):
        url = '/marvin/api/cubes/8485-1901/'
        public = sas == 'public'
        test = sas == 'test'
        if sas == 'testpub':
            public = test = True
        sas = 'utah' if sas != 'local' else sas
        config.switchSasUrl(sas, public=public, test=test)

        e = urlsplit(config.sasurl)
        t = urlsplit(url)
        final = urlunsplit(tuple(strjoin(*z) for z in zip(e, t)))
        assert exp == final

    # @pytest.mark.parametrize('set_default, defrel, exp',
    #                          [('MPL-5', 'MPL-5', 'api.sdss.org'),
    #                           ('DR15', 'DR15', 'dr15.sdss.org/api')], indirect=['set_default'])
    # def test_sasurl_default_release(self, set_default, defrel, exp):
    #     assert config.release == defrel
    #     assert exp in config.sasurl
    #


def strjoin(str1, str2):
    ''' joins two url strings '''
    if not str2.startswith(str1):
        f = os.path.join(str1, str2.lstrip('/')) if str2 else str1
    else:
        f = str2
    return f


@pytest.mark.usefixtures('saslocal')
class TestLogin(object):

    def test_login_fail(self, monkeypatch):
        assert config.token is None
        monkeypatch.setattr(config, 'access', 'public')
        with pytest.raises(AssertionError) as cm:
            config.login()
        assert 'You must have collaboration access to login.' in str(cm.value)

    def test_login(self):
        assert config.token is None
        config.login()
        assert config.token is not None






# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-08 16:31:45
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-31 15:39:47

from __future__ import print_function, division, absolute_import
import pytest
from sdss_access.path import Path


class TestPath(object):

    def test_missing_keys(self, path):
        with pytest.raises(KeyError) as cm:
            full = path.full('mangacube')
        assert 'Missing required keyword arguments:' in str(cm.value)

    def test_public(self):
        path = Path()
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901')
        assert 'mangawork' in url

        release = 'dr14'
        path = Path(public=True, release=release)
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901')
        assert release in url

    @pytest.mark.parametrize('place, exp', [('local', False), ('remote', True)])
    def test_existence(self, path, place, exp):
        full = path.full('mangaimage', drpver='v2_4_3', plate=8116, ifu=1901, dir3d='mastar')
        exists = path.exists('', full=full, remote=(place == 'remote'))
        assert exp == exists

